import statistics
import time

from queue import Empty, Queue
from threading import Thread
from typing import Iterable, Optional, Union

from platypush.message.event.adafruit import (
    AdafruitConnectedEvent,
    AdafruitDisconnectedEvent,
    AdafruitFeedUpdateEvent,
)
from platypush.plugins import RunnablePlugin, action


class AdafruitIoPlugin(RunnablePlugin):
    """
    This plugin allows you to interact with `Adafruit IO
    <https://io.adafruit.com>`_, a cloud-based message queue and storage. You
    can use this plugin to send and receive data to topics connected to your
    Adafruit IO account.

    Some example usages:

      .. code-block:: javascript

        // Send the temperature value for a connected sensor to the "temperature" feed
        {
            "type": "request",
            "action": "adafruit.io.send",
            "args": {
                "feed": "temperature",
                "value": 25.0
            }
        }

        // Receive the most recent temperature value
        {
            "type": "request",
            "action": "adafruit.io.receive",
            "args": {
                "feed": "temperature"
            }
        }

    """

    def __init__(
        self,
        username: str,
        key: str,
        feeds: Iterable[str] = (),
        throttle_interval: Optional[float] = None,
        **kwargs
    ):
        """
        :param username: Your Adafruit username
        :param key: Your Adafruit IO key
        :param feeds: List of feeds to subscribe to. If not set, then the plugin
            will not subscribe to any feed unless instructed to do so, neither
            it will emit any event.
        :param throttle_interval: If set, then instead of sending the values
            directly over ``send`` the plugin will first collect all the
            samples within the specified period and then dispatch them to
            Adafruit IO. You may want to set it if you have data sources
            providing a lot of data points and you don't want to hit the
            throttling limitations of Adafruit.
        """

        from Adafruit_IO import Client, MQTTClient

        super().__init__(**kwargs)

        self._username = username
        self._key = key
        self.feeds = feeds
        self.aio = Client(username=username, key=key)
        self._mqtt_client: Optional[MQTTClient] = None
        self.throttle_interval = throttle_interval
        self._data_throttler_thread: Optional[Thread] = None
        self._data_throttler_queue = Queue()

    @property
    def _mqtt(self):
        if not self._mqtt_client:
            from Adafruit_IO import MQTTClient

            self._mqtt_client = MQTTClient(self._username, self._key)
            self._mqtt_client.on_connect = self._on_connect  # type: ignore
            self._mqtt_client.on_disconnect = self._on_disconnect  # type: ignore
            self._mqtt_client.on_message = self._on_message  # type: ignore

        return self._mqtt_client

    def _on_connect(self, *_, **__):
        assert self._mqtt_client, 'MQTT client not initialized'
        for feed in self.feeds:
            self._mqtt_client.subscribe(feed)
        self._bus.post(AdafruitConnectedEvent())

    def _on_disconnect(self, *_, **__):
        self._bus.post(AdafruitDisconnectedEvent())

    def _on_message(self, _, feed, data, *__):
        try:
            data = float(data)
        except (TypeError, ValueError):
            pass

        self._bus.post(AdafruitFeedUpdateEvent(feed=feed, data=data))

    def _subscribe(self, feed: str):
        assert self._mqtt_client, 'MQTT client not initialized'
        self._mqtt_client.subscribe(feed)

    def _unsubscribe(self, feed: str):
        assert self._mqtt_client, 'MQTT client not initialized'
        self._mqtt_client.unsubscribe(feed)

    def _data_throttler(self):
        from Adafruit_IO import ThrottlingError

        if not self.throttle_interval:
            return

        last_processed_batch_timestamp = None
        data = {}

        try:
            while not self.should_stop():
                try:
                    new_data = self._data_throttler_queue.get(timeout=1)
                    for key, value in new_data.items():
                        data.setdefault(key, []).append(value)
                except Empty:
                    continue

                should_process = data and (
                    last_processed_batch_timestamp is None
                    or time.time() - last_processed_batch_timestamp
                    >= self.throttle_interval
                )

                if not should_process:
                    continue

                last_processed_batch_timestamp = time.time()
                self.logger.info('Processing feeds batch for Adafruit IO')

                for feed, values in data.items():
                    if values:
                        value = statistics.mean(values)

                        try:
                            self.send(feed, value, enqueue=False)
                        except ThrottlingError:
                            self.logger.warning(
                                'Adafruit IO throttling threshold hit, taking a nap '
                                + 'before retrying'
                            )
                            self.wait_stop(self.throttle_interval)

                data = {}
        except Exception as e:
            self.logger.exception(e)

    @action
    def send(self, feed: str, value: Union[int, float, str], enqueue: bool = True):
        """
        Send a value to an Adafruit IO feed.

        :param feed: Feed name.
        :param value: Value to send.
        :param enqueue: If throttle_interval is set, this method by default will append values to the throttling queue
            to be periodically flushed instead of sending the message directly. In such case, pass enqueue=False to
            override the behaviour and send the message directly instead.
        """

        if not self.throttle_interval or not enqueue:
            # If no throttling is configured, or enqueue is false then send the value directly to Adafruit
            self.aio.send(feed, value)
        else:
            # Otherwise send it to the queue to be picked up by the throttler thread
            self._data_throttler_queue.put({feed: value})

    @action
    def send_location_data(
        self,
        feed: str,
        latitude: float,
        longitude: float,
        elevation: float,
        value: Optional[Union[int, float, str]] = None,
    ):
        """
        Send location data to an Adafruit IO feed

        :param feed: Feed name
        :param lat: Latitude
        :param lon: Longitude
        :param ele: Elevation
        :param value: Extra value to attach to the record
        """

        self.aio.send_data(
            feed=feed,
            value=value,
            metadata={
                'lat': latitude,
                'lon': longitude,
                'ele': elevation,
            },
        )

    @action
    def subscribe(self, feed: str):
        """
        Subscribe to a feed.

        :param feed: Feed name
        """
        self._subscribe(feed)

    @action
    def unsubscribe(self, feed: str):
        """
        Unsubscribe from a feed.

        :param feed: Feed name
        """
        self._unsubscribe(feed)

    @classmethod
    def _cast_value(cls, value):
        try:
            value = float(value)
        except ValueError:
            pass

        return value

    def _convert_data_to_dict(self, *data):
        from Adafruit_IO.model import DATA_FIELDS

        return [
            {
                attr: self._cast_value(getattr(i, attr))
                if attr == 'value'
                else getattr(i, attr)
                for attr in DATA_FIELDS
                if getattr(i, attr) is not None
            }
            for i in data
        ]

    @action
    def receive(self, feed: str, limit: int = 1):
        """
        Receive data from the specified Adafruit IO feed

        :param feed: Feed name
        :param limit: Maximum number of data points to be returned. If None,
            all the values in the feed will be returned. Default: 1 (return most
            recent value)
        """

        if limit == 1:
            values = self._convert_data_to_dict(self.aio.receive(feed))
            return values[0] if values else None

        values = self._convert_data_to_dict(*self.aio.data(feed))
        return values[:limit] if limit else values

    @action
    def receive_next(self, feed: str):
        """
        Receive the next unprocessed data point from a feed

        :param feed: Feed name
        """

        values = self._convert_data_to_dict(self.aio.receive_next(feed))
        return values[0] if values else None

    @action
    def receive_previous(self, feed: str):
        """
        Receive the last processed data point from a feed

        :param feed: Feed name
        """

        values = self._convert_data_to_dict(self.aio.receive_previous(feed))
        return values[0] if values else None

    @action
    def delete(self, feed: str, data_id: str):
        """
        Delete a data point from a feed

        :param feed: Feed name
        :param data_id: Data point ID to remove
        """

        self.aio.delete(feed, data_id)

    def main(self):
        if self.throttle_interval:
            self._data_throttler_thread = Thread(target=self._data_throttler)
            self._data_throttler_thread.start()

        try:
            while not self.should_stop():
                cur_wait = 1
                max_wait = 60

                try:
                    self._mqtt.connect()
                    cur_wait = 1
                    self._mqtt.loop_blocking()
                except Exception as e:
                    self.logger.warning(
                        'Adafruit IO connection error: %s, retrying in %d seconds',
                        e,
                        cur_wait,
                    )

                    self.logger.exception(e)
                    self.wait_stop(cur_wait)
                    cur_wait = min(cur_wait * 2, max_wait)
                finally:
                    self._stop_mqtt()
        finally:
            self._stop_data_throttler()

    def _stop_mqtt(self):
        if self._mqtt_client:
            self._mqtt_client.disconnect()
            self._mqtt_client = None

    def _stop_data_throttler(self):
        if self._data_throttler_thread:
            self._data_throttler_thread.join()
            self._data_throttler_thread = None

    def _stop(self):
        self._stop_mqtt()
        self._stop_data_throttler()

    def stop(self):
        self._stop()
        super().stop()


# vim:sw=4:ts=4:et:



import base64
from collections import namedtuple
import json
from typing import Dict, List, Optional, Union
import threading

from serial import Serial

from platypush.common.sensors import Numeric
from platypush.entities.devices import Device
from platypush.entities.sensors import RawSensor, NumericSensor
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin
from platypush.utils import get_lock

_DeviceAndRate = namedtuple('_DeviceAndRate', ['device', 'baud_rate'])


class SerialPlugin(SensorPlugin):
    """
    The serial plugin can read data from a serial device.

    If the device returns a JSON string, then that string will be parsed for
    individual values. For example:

        .. code-block:: json

            {"temperature": 25.0, "humidity": 15.0}

    If the serial device returns such a value, then ``temperature`` and
    ``humidity`` will be parsed as separate entities with the same names as the
    keys provided on the payload.

    The JSON option is a good choice if you have an Arduino/ESP-like device
    whose code you can control, as it allows to easily send data to Platypush
    in a simple key-value format. The use-case would be that of an Arduino/ESP
    device that pushes data on the wire, and this integration would then listen
    for updates.

    Alternatively, you can also use this integration in a more traditional way
    through the :meth:`.read` and :meth:`.write` methods to read and write data
    to the device. In such a case, you may want to disable the "smart polling"
    by setting ``enable_polling`` to ``False`` in the configuration.

    If you want an out-of-the-box solution with a Firmata-compatible firmware,
    you may consider using the :class:`platypush.plugin.arduino.ArduinoPlugin`
    instead.

    Note that device paths on Linux may be subject to change. If you want to
    create static naming associations for your devices (e.g. make sure that
    your Arduino will always be symlinked to ``/dev/arduino`` instead of
    ``/dev/ttyUSB<n>``), you may consider creating `static mappings through
    udev
    <https://dev.to/enbis/how-udev-rules-can-help-us-to-recognize-a-usb-to-serial-device-over-dev-tty-interface-pbk>`_.
    """

    _default_lock_timeout: float = 2.0

    def __init__(
        self,
        device: str,
        baud_rate: int = 9600,
        max_size: int = 1 << 19,
        timeout: float = _default_lock_timeout,
        enable_polling: bool = True,
        poll_interval: float = 0.1,
        **kwargs,
    ):
        """
        :param device: Device path (e.g. ``/dev/ttyUSB0`` or ``/dev/ttyACM0``).
        :param baud_rate: Serial baud rate (default: 9600)
        :param max_size: Maximum size of a JSON payload (default: 512 KB). The
            plugin will keep reading bytes from the wire until it can form a
            valid JSON payload, so this upper limit is required to prevent the
            integration from listening forever and dumping garbage in memory.
        :param timeout: This integration will ensure that only one
            reader/writer can access the serial device at the time, in order to
            prevent mixing up bytes in the response. This value specifies how
            long we should wait for a pending action to terminate when we try
            to run a new action. Default: 2 seconds.
        :param enable_polling: If ``False``, the plugin will not poll the
            device for updates. This can be the case if you want to
            programmatically interface with the device via the :meth:`.read`
            and :meth:`.write` methods instead of polling for updates in JSON
            format.
        :param poll_interval: How often (in seconds) we should poll the device
            for new data. Since we are reading JSON data from a serial
            interface whenever it's ready, the default here can be quite low
            (default: 0.1 seconds).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)

        self.device = device
        self.baud_rate = baud_rate
        self.serial = None
        self.serial_lock = threading.RLock()
        self.last_data: dict = {}
        self._max_size = max_size
        self._timeout = timeout
        self._enable_polling = enable_polling

    def _read_json(
        self,
        serial_port: Serial,
        max_size: Optional[int] = None,
    ) -> str:
        """
        Reads a JSON payload from the wire. It counts the number of curly
        brackets detected, ignoring everything before the first curly bracket,
        and it stops when the processed payload has balanced curly brackets -
        i.e. it can be mapped to JSON.

        :param serial_port: Serial connection.
        :param max_size: Default ``max_size`` override.
        """
        n_brackets = 0
        is_escaped_ch = False
        parse_start = False
        output = bytes()
        max_size = max_size or self._max_size

        while True:
            assert len(output) <= max_size, (
                'Maximum allowed size exceeded while reading from the device: '
                f'read {len(output)} bytes'
            )

            ch = serial_port.read()
            if not ch:
                break

            try:
                ch.decode()
            except Exception as e:
                self.logger.warning('Could not decode character: {}'.format(str(e)))
                output = bytes()

            if ch.decode() == '{' and not is_escaped_ch:
                parse_start = True
                n_brackets += 1

            if not parse_start:
                continue

            output += ch

            if ch.decode() == '}' and not is_escaped_ch:
                n_brackets -= 1
                if n_brackets == 0:
                    break

            is_escaped_ch = ch.decode() == '\\'

        return output.decode().strip()

    def __get_serial(
        self,
        device: Optional[str] = None,
        baud_rate: Optional[int] = None,
        reset: bool = False,
    ) -> Serial:
        """
        Return a ``Serial`` connection object to the given device.

        :param device: Default device path override.
        :param baud_rate: Default baud rate override.
        :param reset: By default, if a connection to the device is already open
            then the current object will be returned. If ``reset=True``, the
            connection will be reset and a new one will be created instead.
        """
        if not device:
            assert self.device, 'No device specified nor default device configured'
            device = self.device

        if baud_rate is None:
            assert self.baud_rate, 'No baud_rate specified nor default configured'
            baud_rate = self.baud_rate

        if self.serial:
            if not reset:
                return self.serial

            self._close_serial()

        self.serial = Serial(device, baud_rate)
        return self.serial

    def _get_serial(
        self,
        device: Optional[str] = None,
        baud_rate: Optional[int] = None,
    ) -> Serial:
        """
        Return a ``Serial`` connection object to the given device.

        :param device: Default device path override.
        :param baud_rate: Default baud rate override.
        """
        try:
            return self.__get_serial(device, baud_rate)
        except AssertionError as e:
            raise e
        except Exception as e:
            self.logger.debug(e)
            self.wait_stop(1)
            return self.__get_serial(device=device, baud_rate=baud_rate, reset=True)

    def _close_serial(self):
        """
        Close the serial connection if it's currently open.
        """
        if self.serial:
            try:
                self.serial.close()
            except Exception as e:
                self.logger.warning('Error while closing serial communication: %s', e)
                self.logger.exception(e)

            self.serial = None

    def _get_device_and_baud_rate(
        self, device: Optional[str] = None, baud_rate: Optional[int] = None
    ) -> _DeviceAndRate:
        """
        Gets the device path and baud rate from the given device and baud rate
        if set, or it falls back on the default configured ones.

        :raise AssertionError: If neither ``device`` nor ``baud_rate`` is set
            nor configured.
        """

        if not device:
            assert (
                self.device
            ), 'No device specified and a default one is not configured'
            device = self.device

        if baud_rate is None:
            assert (
                self.baud_rate is not None
            ), 'No baud_rate specified nor a default value is configured'
            baud_rate = self.baud_rate

        return _DeviceAndRate(device, baud_rate)

    @action
    def get_measurement(
        self,
        *_,
        device: Optional[str] = None,
        baud_rate: Optional[int] = None,
        **__,
    ) -> Dict[str, Numeric]:
        """
        Reads JSON data from the serial device and returns it as a message

        :param device: Device path (default: default configured device).
        :param baud_rate: Baud rate (default: default configured baud_rate).
        """

        device, baud_rate = self._get_device_and_baud_rate(device, baud_rate)

        with get_lock(self.serial_lock, timeout=self._timeout) as serial_available:
            if serial_available:
                ser = self._get_serial(device=device, baud_rate=baud_rate)
                data = self._read_json(ser)

                try:
                    data = dict(json.loads(data))
                except (ValueError, TypeError) as e:
                    raise AssertionError(
                        f'Invalid JSON message from {device}: {e}. Message: {data}'
                    ) from e
            else:
                data = self.last_data

        if data:
            self.last_data = data

        return data

    @action
    def read(
        self,
        device: Optional[str] = None,
        baud_rate: Optional[int] = None,
        size: Optional[int] = None,
        end: Optional[Union[int, str]] = None,
        binary: bool = False,
    ) -> str:
        """
        Reads raw data from the serial device

        :param device: Device to read (default: default configured device)
        :param baud_rate: Baud rate (default: default configured baud_rate)
        :param size: Number of bytes to read
        :param end: End of message, as a character or bytecode
        :param binary: If set to ``True``, then the serial output will be
            interpreted as binary data and a base64-encoded representation will
            be returned. Otherwise, the output will be interpreted as a UTF-8
            encoded string.
        :return: The read message as a UTF-8 string if ``binary=False``,
            otherwise as a base64-encoded string.
        """

        device, baud_rate = self._get_device_and_baud_rate(device, baud_rate)
        assert not (
            (size is None and end is None) or (size is not None and end is not None)
        ), 'Either size or end must be specified'

        assert not (
            end and isinstance(end, str) and len(end) > 1
        ), 'The serial end must be a single character, not a string'

        data = bytes()

        with get_lock(self.serial_lock, timeout=self._timeout) as serial_available:
            assert serial_available, 'Serial read timed out'

            ser = self._get_serial(device=device, baud_rate=baud_rate)
            if size is not None:
                for _ in range(0, size):
                    data += ser.read()
            elif end is not None:
                end_byte = end.encode() if isinstance(end, str) else bytes([end])
                ch = None

                while ch != end_byte:
                    ch = ser.read()
                    if ch != end:
                        data += ch

        return base64.b64encode(data).decode() if binary else data.decode('utf-8')

    @action
    def write(
        self,
        data: Union[str, dict, list, bytes, bytearray],
        device: Optional[str] = None,
        baud_rate: Optional[int] = None,
        binary: bool = False,
    ):
        """
        Writes data to the serial device.

        :param device: Device to write (default: default configured device).
        :param baud_rate: Baud rate (default: default configured baud_rate).
        :param data: Data to send to the serial device. It can be any of the following:

            - A UTF-8 string
            - A base64-encoded string (if ``binary=True``)
            - A dictionary/list that will be encoded as JSON
            - A bytes/bytearray sequence

        :param binary: If ``True``, then the message is either a
            bytes/bytearray sequence or a base64-encoded string.
        """

        device, baud_rate = self._get_device_and_baud_rate(device, baud_rate)
        if isinstance(data, (dict, list)):
            data = json.dumps(data)
        if isinstance(data, str):
            data = base64.b64decode(data) if binary else data.encode('utf-8')

        data = bytes(data)
        with get_lock(self.serial_lock, timeout=self._timeout) as serial_available:
            assert serial_available, 'Could not acquire the device lock'
            ser = self._get_serial(device=device, baud_rate=baud_rate)
            self.logger.info('Writing %d bytes to %s', len(data), device)
            ser.write(data)

    def transform_entities(self, entities: Dict[str, Numeric]) -> List[Device]:
        transformed_entities = []

        for k, v in entities.items():
            sensor_id = f'serial:{k}'
            try:
                value = float(v)
                entity_type = NumericSensor
            except (TypeError, ValueError):
                value = v
                entity_type = RawSensor

            transformed_entities.append(
                entity_type(
                    id=sensor_id,
                    name=k,
                    value=value,
                )
            )

        return [
            Device(
                id='serial',
                name=self.device,
                children=transformed_entities,
            )
        ]

    def main(self):
        if not self._enable_polling:
            # If the polling is disabled, we don't need to do anything here
            self.wait_stop()
            return

        super().main()

    def stop(self):
        super().stop()
        self._close_serial()


# vim:sw=4:ts=4:et:



import datetime
import os
import re

from dataclasses import dataclass
from logging import getLogger
from threading import RLock
from typing import Optional, Iterable

from platypush.plugins.file.monitor import (
    FileMonitorPlugin,
    EventHandler,
    MonitoredResource,
)
from platypush.context import get_bus
from platypush.message.event.log.http import HttpLogEvent

logger = getLogger(__name__)


class LogEventHandler(EventHandler):
    http_line_regex = re.compile(
        r'^([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+\[([^]]+)]\s+"([^"]+)"\s+([\d]+)\s+'
        r'([\d]+)\s*("([^"\s]+)")?\s*("([^"]+)")?$'
    )

    @dataclass
    class FileResource:
        path: str
        pos: int = 0
        lock: RLock = RLock()
        last_timestamp: Optional[datetime.datetime] = None

    def __init__(
        self, *args, monitored_files: Optional[Iterable[str]] = None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._monitored_files = {}
        self.monitor_files(monitored_files or [])

    def monitor_files(self, files: Iterable[str]):
        self._monitored_files.update(
            {f: self.FileResource(path=f, pos=self._get_size(f)) for f in files}
        )

    @staticmethod
    def _get_size(file: str) -> int:
        try:
            return os.path.getsize(file)
        except Exception as e:
            logger.warning('Could not open {}: {}'.format(file, str(e)))
            return 0

    def on_created(self, event):
        self._reset_file(event.src_path)

    def on_deleted(self, event):
        self._reset_file(event.src_path)

    def on_moved(self, event):
        self._reset_file(event.src_path)

    def _reset_file(self, path: str):
        file_info = self._monitored_files.get(path)
        if not file_info:
            return

        file_info.pos = 0

    def on_modified(self, event):
        file_info = self._monitored_files.get(event.src_path)
        if not file_info:
            return

        try:
            file_size = os.path.getsize(event.src_path)
        except OSError as e:
            logger.warning(
                'Could not get the size of {}: {}'.format(event.src_path, str(e))
            )
            return

        if file_info.pos > file_size:
            logger.warning(
                'The size of {} been unexpectedly decreased from {} to {} bytes'.format(
                    event.src_path, file_info.pos, file_size
                )
            )
            file_info.pos = 0

        try:
            with file_info.lock, open(event.src_path, 'r') as f:
                f.seek(file_info.pos)
                for line in f.readlines():
                    evt = self._build_event(file=event.src_path, line=line)
                    if evt and (
                        not file_info.last_timestamp
                        or evt.args['time'] >= file_info.last_timestamp
                    ):
                        get_bus().post(evt)
                        file_info.last_timestamp = evt.args['time']

                file_info.pos = f.tell()
        except OSError as e:
            logger.warning(
                'Error while reading from {}: {}'.format(self.resource.path, str(e))
            )

    @classmethod
    def _build_event(cls, file: str, line: str) -> Optional[HttpLogEvent]:
        line = line.strip()
        if not line:
            return

        m = cls.http_line_regex.match(line.strip())
        if not m:
            logger.warning('Could not parse log line from {}: {}'.format(file, line))
            return

        url = None
        method = 'GET'
        http_version = '1.0'

        try:
            url = m.group(5).split(' ')[1]
            method = m.group(5).split(' ')[0]
            http_version = m.group(5).split(' ')[2].split('/')[1]
        except Exception as e:
            logger.debug(str(e))

        if not url:
            return

        info = {
            'address': m.group(1),
            'user_identifier': m.group(2),
            'user_id': m.group(3),
            'time': datetime.datetime.strptime(m.group(4), '%d/%b/%Y:%H:%M:%S %z'),
            'method': method,
            'url': url,
            'http_version': http_version,
            'status': int(m.group(6)),
            'size': int(m.group(7)),
            'referrer': m.group(9),
            'user_agent': m.group(11),
        }

        for attr, value in info.items():
            if value == '-':
                info[attr] = None

        return HttpLogEvent(logfile=file, **info)


class LogHttpPlugin(FileMonitorPlugin):
    """
    This plugin can be used to monitor one or more HTTP log files (tested on
    Apache and Nginx) and trigger events whenever a new log line is added.
    """

    def __init__(
        self, paths: Iterable[str], log_files: Optional[Iterable[str]] = None, **kwargs
    ):
        """
        :param paths: List of log files to be monitored.
        """
        if log_files:
            self.logger.warning(
                'The log_files parameter is deprecated, use paths instead'
            )

        paths = {os.path.expanduser(log) for log in {*paths, *(log_files or [])}}
        directories = {os.path.dirname(log) for log in paths}
        super().__init__(paths=directories, **kwargs)

        handlers = self._observer._handlers
        for hndls in handlers.values():
            for hndl in hndls:
                hndl.monitor_files(paths)

    @staticmethod
    def event_handler_from_resource(resource: str) -> LogEventHandler:
        return LogEventHandler.from_resource(MonitoredResource(resource))



import json
import os
import requests

from requests.auth import HTTPBasicAuth
from typing import Optional, Sequence, Union, Dict, List, Any

from platypush.plugins import Plugin, action
from platypush.schemas.camera.android.ipcam import CameraStatusSchema


class AndroidIpcam:
    """
    IPCam camera configuration.
    """

    args = {}

    def __init__(
        self,
        name: str,
        host: str,
        port: int = 8080,
        username: Optional[str] = None,
        password: Optional[str] = None,
        timeout: int = 10,
        ssl: bool = True,
    ):
        self.args = {
            'name': name,
            'host': host,
            'port': port,
            'username': username,
            'password': password,
            'timeout': timeout,
            'ssl': ssl,
        }

        self.auth = None
        if username:
            self.auth = HTTPBasicAuth(self.username, self.password)

    def __getattr__(self, item):
        if item in self.args:
            return super().__getattribute__('args').get(item)
        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        if key not in self.args:
            super().__setattr__(key, value)
        else:
            self.args[key] = value

    def __str__(self):
        return json.dumps(self.args or {})

    @property
    def base_url(self) -> str:
        return 'http{ssl}://{host}:{port}/'.format(
            ssl=('s' if self.ssl else ''), host=self.host, port=self.port
        )

    @property
    def stream_url(self) -> str:
        return self.base_url + 'video'

    @property
    def audio_url(self) -> str:
        return self.base_url + 'audio.wav'

    @property
    def image_url(self) -> str:
        return self.base_url + 'photo.jpg'


class CameraAndroidIpcamPlugin(Plugin):
    """
    Plugin to control remote Android cameras over
    `IPCam <https://play.google.com/store/apps/details?id=com.pas.webcam>`_.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        host: Optional[str] = None,
        port: int = 8080,
        username: Optional[str] = None,
        password: Optional[str] = None,
        timeout: int = 10,
        ssl: bool = True,
        cameras: Optional[Sequence[dict]] = None,
        **kwargs,
    ):
        """
        :param name: Custom name for the default camera (default: IP/hostname)
        :param host: Camera host name or address
        :param port: Camera port
        :param username: Camera username, if set
        :param password: Camera password, if set
        :param timeout: Connection timeout
        :param ssl: Use HTTPS instead of HTTP
        :param cameras: Alternatively, you can specify a list of IPCam cameras as a
            list of objects with ``name``, ``host``, ``port``, ``username``,
            ``password``, ``timeout`` and ``ssl`` attributes.
        """
        super().__init__(**kwargs)
        self.cameras: List[AndroidIpcam] = []
        self._camera_name_to_idx: Dict[str, int] = {}

        if not cameras:
            assert host, 'You need to specify at least one camera'
            name = name or host
            camera = AndroidIpcam(
                name=name,
                host=host,
                port=port,
                username=username,
                password=password,
                timeout=timeout,
                ssl=ssl,
            )
            self.cameras.append(camera)
            self._camera_name_to_idx[name] = 0
        else:
            for camera in cameras:
                assert 'host' in camera, 'You need to specify the host for each camera'
                name = camera.get('name', camera['host'])
                camera = AndroidIpcam(
                    name=name,
                    host=camera['host'],
                    port=camera.get('port', port),
                    username=camera.get('username'),
                    password=camera.get('password'),
                    timeout=camera.get('timeout', timeout),
                    ssl=camera.get('ssl', ssl),
                )
                self._camera_name_to_idx[name] = len(self.cameras)
                self.cameras.append(camera)

    def _get_camera(self, camera: Optional[Union[int, str]] = None) -> AndroidIpcam:
        if not camera:
            camera = 0

        if isinstance(camera, int):
            return self.cameras[camera]

        return self.cameras[self._camera_name_to_idx[camera]]

    def _exec(
        self, url: str, *args, camera: Optional[Union[int, str]] = None, **kwargs
    ) -> Union[Dict[str, Any], bool]:
        cam = self._get_camera(camera)
        url = cam.base_url + url
        response = requests.get(
            url, auth=cam.auth, timeout=cam.timeout, verify=False, *args, **kwargs
        )
        response.raise_for_status()

        if response.headers.get('content-type') == 'application/json':
            return response.json()

        return response.text.find('Ok') != -1

    def _change_setting(
        self,
        key: str,
        value: Union[str, int, bool],
        camera: Optional[Union[int, str]] = None,
    ) -> bool:
        if isinstance(value, bool):
            payload = "on" if value else "off"
        else:
            payload = value

        return bool(
            self._exec(
                "settings/{key}?set={payload}".format(key=key, payload=payload),
                camera=camera,
            )
        )

    @action
    def change_setting(
        self,
        key: str,
        value: Union[str, int, bool],
        camera: Optional[Union[int, str]] = None,
    ) -> bool:
        """
        Change a setting.

        :param key: Setting name
        :param value: Setting value
        :param camera: Camera index or configured name
        :return: True on success, False otherwise
        """
        return self._change_setting(key, value, camera=camera)

    @action
    def status(self, camera: Optional[Union[int, str]] = None) -> List[dict]:
        """
        :param camera: Camera index or name (default: status of all the cameras)
        :return: .. schema:: camera.android.ipcam.CameraStatusSchema(many=True)
        """
        cameras = self._camera_name_to_idx.keys() if camera is None else [camera]
        statuses = []

        for c in cameras:
            try:
                print('****** HERE ******')
                print(self._camera_name_to_idx)
                if isinstance(c, int):
                    cam = self.cameras[c]
                else:
                    cam = self.cameras[self._camera_name_to_idx[c]]

                response = self._exec(
                    'status.json', params={'show_avail': 1}, camera=cam.name
                )
                assert isinstance(response, dict), f'Invalid response: {response}'

                status_data = response.get('curvals', {})
                status = CameraStatusSchema().dump(
                    {
                        'name': cam.name,
                        'stream_url': cam.stream_url,
                        'image_url': cam.image_url,
                        'audio_url': cam.audio_url,
                        **status_data,
                    }
                )

                statuses.append(status)
            except Exception as e:
                self.logger.warning(
                    'Could not get the status of %s: %s: %s', c, type(e), e
                )

        return statuses

    @action
    def set_front_facing_camera(
        self, activate: bool = True, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Enable/disable the front-facing camera."""
        return self._change_setting('ffc', activate, camera=camera)

    @action
    def set_torch(
        self, activate: bool = True, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Enable/disable the torch."""
        url = 'enabletorch' if activate else 'disabletorch'
        return bool(self._exec(url, camera=camera))

    @action
    def set_focus(
        self, activate: bool = True, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Enable/disable the focus."""
        url = 'focus' if activate else 'nofocus'
        return bool(self._exec(url, camera=camera))

    @action
    def start_recording(
        self, tag: Optional[str] = None, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Start recording."""
        params = {
            'force': 1,
            **({'tag': tag} if tag else {}),
        }
        return bool(self._exec('startvideo', params=params, camera=camera))

    @action
    def stop_recording(self, camera: Optional[Union[int, str]] = None) -> bool:
        """Stop recording."""
        return bool(self._exec('stopvideo', params={'force': 1}, camera=camera))

    @action
    def take_picture(
        self, image_file: str, camera: Optional[Union[int, str]] = None
    ) -> dict:
        """
        Take a picture and save it on the local device.

        :return: dict

          .. code-block:: json

            {
              "image_file": "/path/to/image.jpg"
            }

        """
        cam = self._get_camera(camera)
        image_file = os.path.abspath(os.path.expanduser(image_file))
        os.makedirs(os.path.dirname(image_file), exist_ok=True)
        response = requests.get(cam.image_url, auth=cam.auth, verify=False)
        response.raise_for_status()

        with open(image_file, 'wb') as f:
            f.write(response.content)

        return {'image_file': image_file}

    @action
    def set_night_vision(
        self, activate: bool = True, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Enable/disable night vision."""
        return self._change_setting('night_vision', activate, camera=camera)

    @action
    def set_overlay(
        self, activate: bool = True, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Enable/disable video overlay."""
        return self._change_setting('overlay', activate, camera=camera)

    @action
    def set_gps(
        self, activate: bool = True, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Enable/disable GPS."""
        return self._change_setting('gps_active', activate, camera=camera)

    @action
    def set_quality(
        self, quality: int = 100, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Set video quality."""
        return self._change_setting('quality', int(quality), camera=camera)

    @action
    def set_motion_detect(
        self, activate: bool = True, camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Enable/disable motion detect."""
        return self._change_setting('motion_detect', activate, camera=camera)

    @action
    def set_orientation(
        self, orientation: str = 'landscape', camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Set video orientation."""
        return self._change_setting('orientation', orientation, camera=camera)

    @action
    def set_zoom(self, zoom: float, camera: Optional[Union[int, str]] = None) -> bool:
        """Set the zoom level."""
        return bool(
            self._exec('settings/ptz', params={'zoom': float(zoom)}, camera=camera)
        )

    @action
    def set_scenemode(
        self, scenemode: str = 'auto', camera: Optional[Union[int, str]] = None
    ) -> bool:
        """Set video orientation."""
        return self._change_setting('scenemode', scenemode, camera=camera)


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod

from . import EntityManager


class AssistantEntityManager(EntityManager, ABC):
    """
    Base class for voice assistant integrations that support entity management.
    """

    @abstractmethod
    def start_conversation(self, *args, **kwargs):
        """
        Programmatically starts a conversation.
        """
        raise NotImplementedError()

    @abstractmethod
    def stop_conversation(self, *args, **kwargs):
        """
        Programmatically stops a conversation.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_muted(self, *args, **kwargs) -> bool:
        """
        :return: True if the microphone is muted, False otherwise.
        """
        raise NotImplementedError()

    @abstractmethod
    def mute(self, *args, **kwargs):
        """
        Mute the microphone.
        """
        raise NotImplementedError()

    @abstractmethod
    def unmute(self, *args, **kwargs):
        """
        Unmute the microphone.
        """
        raise NotImplementedError()

    def toggle_mute(self, *_, **__):
        """
        Toggle the mute state of the microphone.
        """
        return self.mute() if self.is_muted() else self.unmute()

    @abstractmethod
    def pause_detection(self, *args, **kwargs):
        """
        Put the assistant on pause. No new conversation events will be triggered.
        """
        raise NotImplementedError()

    @abstractmethod
    def resume_detection(self, *args, **kwargs):
        """
        Resume the assistant hotword detection from a paused state.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_detecting(self, *args, **kwargs) -> bool:
        """
        :return: True if the asistant is detecting, False otherwise.
        """
        raise NotImplementedError()

    @abstractmethod
    def send_text_query(self, *args, **kwargs):
        """
        Send a text query to the assistant.
        """
        raise NotImplementedError()



import os
from typing import List, Optional

from .._model import TorrentSearchResult
from ._base import TorrentsCsvBaseProvider
from ._constants import (
    TORRENT_CSV_API_URL,
    TORRENTS_CSV_DOWNLOAD_URL,
    TORRENTS_CSV_DEFAULT_CHECK_INTERVAL,
)
from .api import TorrentsCsvAPIProvider
from .local import TorrentsCsvLocalProvider


class TorrentsCsvSearchProvider(TorrentsCsvBaseProvider):
    """
    Torrent that uses `Torrents.csv <https://torrents-csv.com/>`_ to search
    for torrents, either by using the API or by leveraging a local database.
    """

    def __init__(
        self,
        api_url: str = TORRENT_CSV_API_URL,
        csv_url: str = TORRENTS_CSV_DOWNLOAD_URL,
        download_csv: bool = False,
        csv_path: Optional[str] = None,
        db_path: Optional[str] = None,
        csv_url_check_interval: int = TORRENTS_CSV_DEFAULT_CHECK_INTERVAL,
        **kwargs
    ):
        """
        :param api_url: Torrents.csv API URL.
        :param csv_url: Torrents.csv CSV URL.
        :param download_csv: Whether to download the CSV file.
        :param csv_path: Path to the CSV file.
        :param db_path: Path to the SQLite database file.
        :param csv_url_check_interval: Interval to check for CSV updates.
        """
        super().__init__(**kwargs)
        self.api_url = api_url
        self.csv_url = csv_url
        self.download_csv = download_csv
        self.csv_path = os.path.expanduser(csv_path) if csv_path else None
        self.db_path = os.path.expanduser(db_path) if db_path else None
        self.csv_url_check_interval = csv_url_check_interval

    @property
    def _delegate(self) -> TorrentsCsvBaseProvider:
        """
        :return: The provider to delegate the search to.
        """
        if self.download_csv or self.csv_path or self.db_path:
            return TorrentsCsvLocalProvider(
                download_csv=self.download_csv,
                csv_url=self.csv_url,
                csv_path=self.csv_path,
                db_path=self.db_path,
                csv_url_check_interval=self.csv_url_check_interval,
                enabled=True,
            )

        return TorrentsCsvAPIProvider(api_url=self.api_url, enabled=True)

    @classmethod
    def default_enabled(cls) -> bool:
        """
        This provider is enabled by default.
        """
        return True

    def _search(
        self, query: str, *_, limit: int = 25, page: int = 1, **__
    ) -> List[TorrentSearchResult]:
        """
        Perform a search of torrents using the Torrent.csv API.

        :param query: Query string.
        :param limit: Number of results to return (default: 25).
        :param page: Page number (default: 1).
        """
        results = list(self._delegate.search(query=query, limit=limit, page=page))
        for result in results:
            result.provider = self.provider_name()

        return results


# vim:sw=4:ts=4:et:



from typing import Any, Dict, Iterable, List
from platypush.entities.bluetooth import BluetoothService

from ..._model import ServiceClass


# pylint: disable=too-few-public-methods
class BluetoothServicesBuilder:
    """
    Builds a list of :class:`platypush.entities.bluetooth.BluetoothService`
    entities from the list of dictionaries returned by
    ``bluetooth.find_services()``.
    """

    @classmethod
    def build(cls, services: Iterable[Dict[str, Any]]) -> List[BluetoothService]:
        """
        Parse the services exposed by the device from the raw pybluez data.
        """
        parsed_services = {}

        for srv in services:
            service_args = {
                key: srv.get(key) for key in ['name', 'description', 'port', 'protocol']
            }

            classes = srv.get('service-classes', [])
            versions = dict(srv.get('profiles', []))

            for srv_cls in classes:
                uuid = BluetoothService.to_uuid(srv_cls)
                parsed_service = parsed_services[uuid] = BluetoothService(
                    id=f'{srv["host"]}::{srv_cls}',
                    uuid=uuid,
                    version=versions.get(srv_cls),
                    **service_args,
                )

                # Ensure that the service name is always set
                if not parsed_service.name:
                    parsed_service.name = (
                        str(parsed_service.service_class)
                        if parsed_service.service_class != ServiceClass.UNKNOWN
                        else f'[{parsed_service.uuid}]'
                    )

        return list(parsed_services.values())



from dataclasses import asdict
import os
from typing import Optional

from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.media import MediaPlayRequestEvent, MediaVolumeChangedEvent

from platypush.plugins import action
from platypush.plugins.media.gstreamer.model import MediaPipeline


class MediaGstreamerPlugin(MediaPlugin):
    """
    Plugin to play media over GStreamer.
    """

    def __init__(self, sink: Optional[str] = None, *args, **kwargs):
        """
        :param sink: GStreamer audio sink (default: ``None``, automatic).
        """
        super().__init__(*args, **kwargs)
        self.sink = sink
        self._player: Optional[MediaPipeline] = None
        self._resource: Optional[str] = None

    def _allocate_pipeline(self, resource: str) -> MediaPipeline:
        pipeline = MediaPipeline(resource)
        if self.sink:
            sink = pipeline.add_sink(self.sink, sync=False)
            pipeline.link(pipeline.get_source(), sink)

        self._player = pipeline
        self._resource = resource
        return pipeline

    @action
    def play(self, resource: Optional[str] = None, **_):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
        """

        if not resource:
            if self._player:
                self._player.play()
            return

        resource = self._get_resource(resource)
        path = os.path.abspath(os.path.expanduser(resource))
        if os.path.exists(path):
            resource = 'file://' + path

        MediaPipeline.post_event(MediaPlayRequestEvent, resource=resource)
        pipeline = self._allocate_pipeline(resource)
        pipeline.play()
        if self.volume:
            pipeline.set_volume(self.volume / 100.0)

        return self.status()

    @action
    def pause(self):
        """Toggle the paused state"""
        assert self._player, 'No instance is running'
        self._player.pause()
        return self.status()

    @action
    def quit(self):
        """Stop and quit the player (alias for :meth:`.stop`)"""
        self._stop_torrent()
        assert self._player, 'No instance is running'

        self._player.stop()
        self._player = None
        return {'state': PlayerState.STOP.value}

    @action
    def stop(self):
        """Stop and quit the player (alias for :meth:`.quit`)"""
        return self.quit()

    @action
    def voldown(self, step=10.0):
        """Volume down by (default: 10)%"""
        # noinspection PyUnresolvedReferences
        return self.set_volume(self.get_volume().output - step)

    @action
    def volup(self, step=10.0):
        """Volume up by (default: 10)%"""
        # noinspection PyUnresolvedReferences
        return self.set_volume(self.get_volume().output + step)

    @action
    def get_volume(self) -> float:
        """
        Get the volume.

        :return: Volume value between 0 and 100.
        """
        assert self._player, 'No instance is running'
        return self._player.get_volume() * 100.0

    @action
    def set_volume(self, volume):
        """
        Set the volume.

        :param volume: Volume value between 0 and 100.
        """
        assert self._player, 'Player not running'
        # noinspection PyTypeChecker
        volume = max(0, min(1, volume / 100.0))
        self._player.set_volume(volume)
        MediaPipeline.post_event(MediaVolumeChangedEvent, volume=volume * 100)
        return self.status()

    @action
    def seek(self, position: float) -> dict:
        """
        Seek backward/forward by the specified number of seconds.

        :param position: Number of seconds relative to the current cursor.
        """
        assert self._player, 'No instance is running'
        cur_pos = self._player.get_position()
        return self.set_position(cur_pos + position)

    @action
    def back(self, offset=60.0):
        """Back by (default: 60) seconds"""
        return self.seek(-offset)

    @action
    def forward(self, offset=60.0):
        """Forward by (default: 60) seconds"""
        return self.seek(offset)

    @action
    def is_playing(self):
        """
        :returns: True if it's playing, False otherwise
        """
        return self._player and self._player.is_playing()

    @action
    def load(self, resource, **_):
        """
        Load/queue a resource/video to the player (alias for :meth:`.play`).
        """
        return self.play(resource)

    @action
    def mute(self):
        """Toggle mute state"""
        assert self._player, 'No instance is running'
        muted = self._player.is_muted()
        if muted:
            self._player.unmute()
        else:
            self._player.mute()

        return {'muted': self._player.is_muted()}

    @action
    def set_position(self, position: float) -> dict:
        """
        Seek backward/forward to the specified absolute position.

        :param position: Stream position in seconds.
        :return: Player state.
        """
        assert self._player, 'No instance is running'
        self._player.seek(position)
        return self.status()

    @action
    def status(self) -> dict:
        """
        Get the current player state.
        """
        if not self._player:
            return {'state': PlayerState.STOP.value}

        pos = self._player.get_position()
        length = self._player.get_duration()

        status = {
            'duration': length,
            'filename': self._resource[7:]
            if self._resource.startswith('file://')
            else self._resource,
            'mute': self._player.is_muted(),
            'name': self._resource,
            'pause': self._player.is_paused(),
            'percent_pos': pos / length
            if pos is not None and length is not None and pos >= 0 and length > 0
            else 0,
            'position': pos,
            'seekable': length is not None and length > 0,
            'state': self._gst_to_player_state(self._player.get_state()).value,
            'url': self._resource,
            'volume': self._player.get_volume() * 100,
        }

        if self._latest_resource:
            status.update(
                {
                    k: v
                    for k, v in asdict(self._latest_resource).items()
                    if v is not None
                }
            )

        return status

    @staticmethod
    def _gst_to_player_state(state) -> PlayerState:
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        from gi.repository import Gst

        if state == Gst.State.READY:
            return PlayerState.STOP
        if state == Gst.State.PAUSED:
            return PlayerState.PAUSE
        if state == Gst.State.PLAYING:
            return PlayerState.PLAY
        return PlayerState.IDLE

    def toggle_subtitles(self, *_, **__):
        raise NotImplementedError

    def set_subtitles(self, *_, **__):
        raise NotImplementedError

    def remove_subtitles(self, *_, **__):
        raise NotImplementedError


# vim:sw=4:ts=4:et:



import contextlib
import importlib
import inspect
from dataclasses import dataclass
from typing import Type, Optional

from .._serialize import Serializable
from . import Constructor
from .component import Component
from .constants import doc_base_url


@dataclass
class Message(Component, Serializable):
    """
    Represents the metadata of a message type (event or response).
    """

    name: str
    type: Type
    doc: Optional[str] = None
    constructor: Optional[Constructor] = None

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "type": f"{self.type.__module__}.{self.type.__qualname__}",
            "doc": self.doc,
            "doc_url": self.doc_url,
            "args": {
                **(
                    {name: arg.to_dict() for name, arg in self.constructor.args.items()}
                    if self.constructor
                    else {}
                ),
            },
        }

    @classmethod
    def by_name(cls, name: str) -> "Message":
        """
        :param name: Message type name.
        :return: A parsed message class given its type.
        """
        return cls.by_type(cls._get_cls(name))

    @classmethod
    def by_type(cls, type: Type) -> "Message":
        """
        :param type: Message type.
        :return: A parsed message class given its type.
        """
        from platypush.message import Message as MessageClass

        assert issubclass(type, MessageClass), f"Expected a Message class, got {type}"
        obj = cls(
            name=f'{type.__module__}.{type.__qualname__}',
            type=type,
            doc=inspect.getdoc(type),
            constructor=Constructor.parse(type),
        )

        for p_type in inspect.getmro(type)[1:]:
            # Don't go upper in the hierarchy.
            if p_type == type:
                break

            with contextlib.suppress(AssertionError):
                p_obj = cls.by_type(p_type)
                # Merge constructor parameters
                if obj.constructor and p_obj.constructor:
                    cls._merge_params(obj.constructor.args, p_obj.constructor.args)

        return obj

    @property
    def cls(self) -> Type:
        """
        :return: The class of a message.
        """
        return self._get_cls(self.name)

    @staticmethod
    def _get_cls(name: str) -> Type:
        """
        :param name: Full qualified type name, module included.
        :return: The associated class.
        """
        tokens = name.split(".")
        module = importlib.import_module(".".join(tokens[:-1]))
        return getattr(module, tokens[-1])

    @property
    def doc_url(self) -> str:
        """
        :return: URL of the documentation for the message.
        """
        from platypush.message.event import Event
        from platypush.message.response import Response

        if issubclass(self.type, Event):
            section = 'events'
        elif issubclass(self.type, Response):
            section = 'responses'
        else:
            raise AssertionError(f'Unknown message type {self.type}')

        mod_name = '.'.join(self.name.split('.')[3:-1])
        return f"{doc_base_url}/{section}/{mod_name}.html#{self.name}"



from marshmallow import fields
from marshmallow.schema import Schema

from platypush.schemas import DateTime


class RssFeedEntrySchema(Schema):
    feed_title = fields.String(metadata={'description': 'Feed title'})
    feed_url = fields.URL(
        required=True,
        metadata={
            'description': 'URL of the feed',
            'example': 'https://some-website/rss',
        },
    )

    id = fields.String(
        required=True,
        metadata={
            'description': 'Feed entry ID',
            'example': '1234',
        },
    )

    url = fields.URL(
        required=True,
        metadata={
            'description': 'URL of the feed entry',
            'example': 'https://some-website/articles/1234',
        },
    )

    published = DateTime(
        required=True, metadata={'description': 'Entry published time'}
    )

    title = fields.String(metadata={'description': 'Feed entry title'})
    summary = fields.String(metadata={'description': 'Feed entry summary'})
    content = fields.String(metadata={'description': 'Feed entry content'})
    author = fields.String(metadata={'description': 'Feed entry author'})
    tags = fields.List(
        fields.String(),
        metadata={'description': 'Feed entry tags'},
    )



from dataclasses import dataclass, field
from typing import Dict, Optional, Tuple, Union
from threading import RLock

IdType = Union[str, int]
_CacheMap = Dict[IdType, dict]


@dataclass
class Cache:
    """
    Cache class for Z-Wave entities.
    """

    by_id: _CacheMap = field(default_factory=dict)
    by_name: _CacheMap = field(default_factory=dict)
    _lock: RLock = field(default_factory=RLock)

    def add(self, *objs: dict, overwrite: bool = False) -> None:
        """
        Add objects to the cache.
        """
        with self._lock:
            if overwrite:
                self.clear()

            for obj in objs:
                obj_id = obj.get("id")
                if obj_id:
                    self.by_id[obj_id] = obj

                name = self._get_name(obj)
                if name:
                    self.by_name[name] = obj

    def get(
        self,
        obj: Optional[Union[IdType, Tuple[Optional[IdType], Optional[IdType]]]] = None,
        default=None,
    ) -> Optional[dict]:
        """
        Get an object from the cache, by ID or name/label.
        """
        if obj is None:
            return None

        if isinstance(obj, tuple):
            return self.get(obj[0], self.get(obj[1], default))

        return self.by_id.get(obj, self.by_name.get(obj, default))

    def __contains__(self, obj: Optional[IdType]) -> bool:
        """
        Check if an object with the given ID/name is in the cache.
        """
        return obj in self.by_id or obj in self.by_name

    def clear(self) -> None:
        """
        Clear the cache.
        """
        with self._lock:
            self.by_id.clear()
            self.by_name.clear()

    def pop(self, obj_id: IdType, default=None) -> Optional[dict]:
        """
        Remove and return an object from the cache.
        """
        with self._lock:
            obj = self.by_id.pop(obj_id, default)
            if not obj:
                return obj

            name = self._get_name(obj)
            if name:
                self.by_name.pop(name, None)
            return obj

    @staticmethod
    def _get_name(obj: dict) -> Optional[str]:
        """
        @return The name/label of an object.
        """
        return obj.get("name", obj.get("label"))


@dataclass
class State:
    """
    State of the Z-Wave network.
    """

    nodes: Cache = field(default_factory=Cache)
    values: Cache = field(default_factory=Cache)
    scenes: Cache = field(default_factory=Cache)
    groups: Cache = field(default_factory=Cache)



from marshmallow_dataclass import class_schema

from .._base import SystemBaseSchema
from ._base import CpuInfoBaseSchema, CpuTimesBaseSchema
from ._model import CpuFrequency, CpuInfo, CpuStats, CpuTimes


CpuFrequencySchema = class_schema(CpuFrequency, base_schema=SystemBaseSchema)
CpuInfoSchema = class_schema(CpuInfo, base_schema=CpuInfoBaseSchema)
CpuTimesSchema = class_schema(CpuTimes, base_schema=CpuTimesBaseSchema)
CpuStatsSchema = class_schema(CpuStats, base_schema=SystemBaseSchema)



from typing import Iterable

from ._base import AudioConverter, dtype_to_ffmpeg_format


class RawInputAudioConverter(AudioConverter):
    """
    Converts raw audio input to a compressed media format.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._dtype and not self._input_format:
            ffmpeg_format = dtype_to_ffmpeg_format.get(self._dtype)
            assert ffmpeg_format, (
                f'Unsupported data type: {self._dtype}. Supported data types: '
                f'{list(dtype_to_ffmpeg_format.keys())}'
            )

            self._input_format = ffmpeg_format

    @property
    def _input_format_args(self) -> Iterable[str]:
        args = (
            '-ar',
            str(self._sample_rate),
            *self._channel_layout_args,
        )

        if self._input_format:
            args = ('-f', self._input_format) + args

        return args

    @property
    def _output_format_args(self) -> Iterable[str]:
        if not self._output_format:
            return ()

        ffmpeg_args = self._format_to_ffmpeg_args.get(self._output_format)
        assert ffmpeg_args, (
            f'Unsupported output format: {self._output_format}. Supported formats: '
            f'{list(self._format_to_ffmpeg_args.keys())}'
        )

        return ffmpeg_args


# vim:sw=4:ts=4:et:



import asyncio
from typing import Iterable, Optional, Union

import aioxmpp
import aioxmpp.im
import aioxmpp.muc.xso

from platypush.message.event.xmpp import (
    XmppRoomAffiliationChangedEvent,
    XmppRoomInviteEvent,
    XmppRoomInviteAcceptedEvent,
    XmppRoomInviteRejectedEvent,
    XmppRoomEnterEvent,
    XmppRoomExitEvent,
    XmppRoomJoinEvent,
    XmppRoomLeaveEvent,
    XmppRoomMessageReceivedEvent,
    XmppRoomNickChangedEvent,
    XmppRoomPresenceChangedEvent,
    XmppRoomRoleChangedEvent,
    XmppRoomTopicChangedEvent,
)

from .._types import Errors, XmppPresence
from ._base import XmppBaseHandler


class XmppRoomHandler(XmppBaseHandler):
    """
    Handler for XMPP room events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.muc_client: aioxmpp.MUCClient = self._client.summon(aioxmpp.MUCClient)
        self.muc_client.on_muc_invitation.connect(self._on_muc_invitation)  # type: ignore

    async def _restore_state(self):
        if self._loaded_state.rooms:
            await asyncio.gather(
                *[
                    self.join(
                        room_id,
                        nick=room.nick if room.nick else self._jid.localpart,
                        password=room.password,
                    )
                    for room_id, room in self._loaded_state.rooms.items()
                ]
            )

    def restore_state(self):
        self._async_run(self._restore_state, wait_result=False)

    def _on_muc_invitation(
        self,
        _: aioxmpp.stanza.Message,
        room_id: aioxmpp.JID,
        inviter: aioxmpp.JID,
        mode: aioxmpp.im.InviteMode,
        password: Optional[str] = None,
        reason: Optional[str] = None,
        **__,
    ):
        def join():
            assert self._loop, Errors.LOOP
            nick = self._jid.localpart
            self._async_run(
                self.join,
                room_id=jid,
                nick=nick,
                password=password,
                timeout=self.DEFAULT_TIMEOUT,
            )

            self._state.pending_rooms.add(jid)
            self._state.room_invites.pop(jid, None)
            self._post_event(XmppRoomInviteAcceptedEvent, room_id=jid)

        def reject():
            self._state.room_invites.pop(jid, None)
            self._post_event(XmppRoomInviteRejectedEvent, room_id=jid)

        jid = str(room_id)
        invite = self._state.room_invites[jid]
        self._post_user_event(
            XmppRoomInviteEvent,
            room_id=jid,
            user_id=inviter,
            mode=mode.name,
            password=password,
            reason=reason,
        )

        invite.on_accepted = join
        invite.on_rejected = reject
        if self._config.auto_accept_invites:
            invite.accept()

    def _get_occupant_by_jid(
        self, user_id: str, room: aioxmpp.muc.Room
    ) -> aioxmpp.muc.service.Occupant:
        occupant = next(
            iter(
                m
                for m in room.members
                if str(m.conversation_jid) == user_id or str(m.direct_jid) == user_id
            ),
            None,
        )

        assert occupant, Errors.NO_USER
        return occupant

    async def join(
        self,
        room_id: str,
        nick: Optional[str] = None,
        password: Optional[str] = None,
        auto_rejoin: bool = True,
    ):
        address = aioxmpp.JID.fromstr(room_id)
        room, future = self.muc_client.join(
            address,
            nick=nick,
            password=password,
            autorejoin=auto_rejoin,
        )

        await future
        await self._register_room(room)
        return room

    async def leave(self, room_id: str):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED
        await room.leave()
        self._unregister_room(room)

    async def invite(
        self,
        user_id: aioxmpp.JID,
        room_id: str,
        mode: aioxmpp.im.InviteMode = aioxmpp.im.InviteMode.DIRECT,
        text: Optional[str] = None,
    ):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED
        await room.invite(user_id, text=text, mode=mode)

    async def kick(
        self,
        user_id: aioxmpp.JID,
        room_id: str,
        reason: Optional[str] = None,
    ):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED

        occupant = self._get_occupant_by_jid(user_id=str(user_id), room=room)
        await room.kick(occupant, reason=reason)

    async def ban(
        self,
        user_id: aioxmpp.JID,
        room_id: str,
        reason: Optional[str] = None,
    ):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED

        occupant = self._get_occupant_by_jid(user_id=str(user_id), room=room)
        await room.ban(occupant, reason=reason)

    async def set_affiliation(
        self,
        user_id: aioxmpp.JID,
        room_id: str,
        affiliation: str,
        reason: Optional[str] = None,
    ):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED

        occupant = self._get_occupant_by_jid(user_id=str(user_id), room=room)
        await room.muc_set_affiliation(
            occupant.direct_jid or occupant.conversation_jid, affiliation, reason=reason
        )

    async def set_role(
        self,
        user_id: aioxmpp.JID,
        room_id: str,
        role: str,
        reason: Optional[str] = None,
    ):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED

        occupant = self._get_occupant_by_jid(user_id=str(user_id), room=room)
        await room.muc_set_role(occupant.nick, role, reason=reason)

    async def set_topic(self, room_id: str, topic: str):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED
        await room.set_topic(topic)

    async def set_nick(self, room_id: str, nick: str):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED
        await room.set_nick(nick)

    # pylint: disable=too-many-branches
    async def set_room_config(
        self,
        room_id: str,
        name: Optional[bool] = None,
        description: Optional[bool] = None,
        members_only: Optional[bool] = None,
        persistent: Optional[bool] = None,
        moderated: Optional[bool] = None,
        allow_invites: Optional[bool] = None,
        allow_private_messages: Optional[bool] = None,
        allow_change_subject: Optional[bool] = None,
        enable_logging: Optional[bool] = None,
        max_history_fetch: Optional[int] = None,
        max_users: Optional[int] = None,
        password_protected: Optional[bool] = None,
        public: Optional[bool] = None,
        room_admins: Optional[Iterable[Union[str, aioxmpp.JID]]] = None,
        room_owners: Optional[Iterable[Union[str, aioxmpp.JID]]] = None,
        password: Optional[str] = None,
        language: Optional[str] = None,
    ):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED
        config = await self.muc_client.get_room_config(room.jid)
        form = aioxmpp.muc.xso.ConfigurationForm.from_xso(config)

        if members_only is not None:
            form.membersonly.value = members_only
        if persistent is not None:
            form.persistentroom.value = persistent
        if moderated is not None:
            form.moderatedroom.value = moderated
        if description is not None:
            form.roomdesc.value = description
        if name is not None:
            form.roomname.value = name
        if allow_invites is not None:
            form.allowinvites.value = allow_invites
        if allow_private_messages is not None:
            form.allowpm.value = allow_private_messages
        if allow_change_subject is not None:
            form.changesubject.value = allow_change_subject
        if enable_logging is not None:
            form.enablelogging.value = enable_logging
        if max_history_fetch is not None:
            form.maxhistoryfetch.value = max_history_fetch
        if max_users is not None:
            form.maxusers.value = max_users
        if password_protected is not None:
            form.passwordprotectedroom.value = max_users
        if public is not None:
            form.publicroom.value = public
        if password is not None:
            form.roomsecret.value = password
        if language is not None:
            form.lang.value = language
        if room_admins is not None:
            form.roomadmins.value = [
                aioxmpp.JID.fromstr(user_id) if isinstance(user_id, str) else user_id
                for user_id in room_admins
            ]
        if room_owners is not None:
            form.roomowners.value = [
                aioxmpp.JID.fromstr(user_id) if isinstance(user_id, str) else user_id
                for user_id in room_owners
            ]

        await self.muc_client.set_room_config(room.jid, form.render_reply())

    async def _register_room(self, room: aioxmpp.muc.Room):
        room_id = str(room.jid)
        if not self._state.rooms.get(room_id):
            self._register_room_events(room)
            if room_id in self._state.pending_rooms:
                self._state.pending_rooms.remove(room_id)

            self._state.rooms[room_id] = room
            self._post_user_room_event(
                XmppRoomJoinEvent,
                room=room,
                user_id=self._jid,
                is_self=True,
                members=[
                    str(m.direct_jid if m.direct_jid else m.conversation_jid)
                    for m in room.members
                ],
            )

            await self._configure_room_on_join(room)

    async def _configure_room_on_join(self, room: aioxmpp.muc.Room):
        # Check if I'm the owner of the room and there's only me here.
        # If that's the case, odds are that the room has been newly created.
        # Newly created rooms have public_room set to False by default
        if len(room.members) != 1:
            return

        member = room.members[0]
        if not (member.is_self and member.affiliation == "owner"):
            return

        config = await self.muc_client.get_room_config(room.jid)
        form = aioxmpp.muc.xso.ConfigurationForm.from_xso(config)

        # If it's already a persistent room, then it's probably not a room that
        # has just been created
        if form.persistentroom.value:
            return

        form.publicroom.value = True
        form.allowinvites.value = True
        await self.muc_client.set_room_config(room.jid, form.render_reply())

    def _unregister_room(self, room: aioxmpp.muc.Room):
        stored_room = self._state.rooms.pop(self._jid_to_str(room.jid), None)
        if stored_room:
            self._post_user_room_event(
                XmppRoomLeaveEvent,
                room=room,
                user_id=self._jid,
                is_self=True,
            )

    def _register_room_events(self, room: aioxmpp.muc.Room):
        room.on_enter.connect(self._on_room_enter(room))  # type: ignore
        room.on_exit.connect(self._on_room_exit(room))  # type: ignore
        room.on_join.connect(self._on_room_join(room))  # type: ignore
        room.on_leave.connect(self._on_room_leave(room))  # type: ignore
        room.on_message.connect(self._on_msg_received(room))  # type: ignore
        room.on_nick_changed.connect(self._on_room_nick_changed(room))  # type: ignore
        room.on_presence_changed.connect(self._on_room_presence_changed(room))  # type: ignore
        room.on_topic_changed.connect(self._on_room_topic_changed(room))  # type: ignore
        room.on_muc_affiliation_changed.connect(self._on_room_muc_affiliation_changed(room))  # type: ignore
        room.on_muc_role_changed.connect(self._on_room_muc_role_changed(room))  # type: ignore

    def _on_msg_received(self, room: aioxmpp.muc.Room):
        def callback(msg, occupant: aioxmpp.muc.service.Occupant, *_, **__):
            if not msg.body:
                return

            if msg.error:
                self.logger.warning(
                    'Error on message from %s: %s', msg.from_, msg.error
                )

            body = msg.body.lookup([aioxmpp.structs.LanguageRange.fromstr('*')])
            self._post_room_occupant_event(
                XmppRoomMessageReceivedEvent,
                room=room,
                occupant=occupant,
                body=body.rstrip(),
            )

        return callback

    def _on_room_join(self, room: aioxmpp.muc.Room):
        def callback(occupant: aioxmpp.muc.service.Occupant, *_, **__):
            self._post_room_occupant_event(
                XmppRoomJoinEvent, room=room, occupant=occupant
            )

        return callback

    def _on_room_leave(self, room: aioxmpp.muc.Room):
        def callback(occupant: aioxmpp.muc.service.Occupant, *_, **__):
            if occupant.is_self:
                self._unregister_room(room)
            else:
                self._post_room_occupant_event(
                    XmppRoomLeaveEvent, room=room, occupant=occupant
                )

        return callback

    def _on_room_enter(self, room: aioxmpp.muc.Room):
        def callback(*args, **__):
            if args:
                occupant = args[0]
                self._post_room_occupant_event(
                    XmppRoomEnterEvent, room=room, occupant=occupant
                )
            else:
                self._async_run(self._register_room, room)
                self._post_user_room_event(
                    XmppRoomEnterEvent,
                    room=room,
                    user_id=self._jid,
                    is_self=True,
                )

        return callback

    def _on_room_exit(self, room: aioxmpp.muc.Room):
        def callback(
            *args,
            reason: Optional[str] = None,
            **__,
        ):
            if args:
                occupant = args[0]
                self._post_room_occupant_event(
                    XmppRoomExitEvent, room=room, occupant=occupant, reason=reason
                )
            else:
                self._post_user_room_event(
                    XmppRoomExitEvent,
                    room=room,
                    user_id=self._jid,
                    is_self=True,
                    reason=reason,
                )

        return callback

    def _on_room_nick_changed(self, room: aioxmpp.muc.Room):
        def callback(
            member: aioxmpp.muc.service.Occupant,
            old_nick: Optional[str],
            new_nick: Optional[str],
            *_,
            **__,
        ):
            self._post_room_occupant_event(
                XmppRoomNickChangedEvent,
                room=room,
                occupant=member,
                old_nick=old_nick,
                new_nick=new_nick,
            )

        return callback

    def _on_room_presence_changed(self, room: aioxmpp.muc.Room):
        def callback(
            occupant: aioxmpp.muc.service.Occupant,
            _,
            presence: aioxmpp.stanza.Presence,
            **__,
        ):
            self._post_room_occupant_event(
                XmppRoomPresenceChangedEvent,
                room=room,
                occupant=occupant,
                status=aioxmpp.PresenceShow(presence.show).value
                or XmppPresence.AVAILABLE.value,
            )

        return callback

    def _on_room_muc_affiliation_changed(self, room: aioxmpp.muc.Room):
        def callback(
            presence: aioxmpp.stanza.Presence,
            *_,
            actor: Optional[aioxmpp.muc.xso.UserActor] = None,
            reason: Optional[str] = None,
            **__,
        ):
            occupant = self._get_occupant_by_jid(room=room, user_id=str(presence.from_))
            self._post_room_occupant_event(
                XmppRoomAffiliationChangedEvent,
                room=room,
                occupant=occupant,
                affiliation=occupant.affiliation,
                changed_by=str(actor.jid) if actor else None,
                reason=reason,
            )

        return callback

    def _on_room_muc_role_changed(self, room: aioxmpp.muc.Room):
        def callback(
            presence: aioxmpp.stanza.Presence,
            *_,
            actor: Optional[aioxmpp.muc.xso.UserActor] = None,
            reason: Optional[str] = None,
            **__,
        ):
            occupant = self._get_occupant_by_jid(room=room, user_id=str(presence.from_))
            self._post_room_occupant_event(
                XmppRoomRoleChangedEvent,
                room=room,
                occupant=occupant,
                role=occupant.role,
                changed_by=str(actor.jid) if actor else None,
                reason=reason,
            )

        return callback

    def _on_room_topic_changed(self, room: aioxmpp.muc.Room):
        def callback(
            member: aioxmpp.muc.service.ServiceMember,
            topic_map: aioxmpp.structs.LanguageMap,
            **_,
        ):
            topic = topic_map.lookup([aioxmpp.structs.LanguageRange.fromstr('*')])
            self._post_room_event(
                XmppRoomTopicChangedEvent,
                room=room,
                topic=topic,
                changed_by=member.nick,
            )

        return callback

    def send_message(
        self,
        room_id: str,
        body: str,
        language: Optional[str] = None,
    ):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED

        target = room.jid
        msg_type = aioxmpp.MessageType.GROUPCHAT
        lang = language or self._lang
        msg = aioxmpp.Message(type_=msg_type, to=target)
        msg.body.update({lang: body})
        self._client.enqueue(msg)

    def accept_invite(self, room_id: str):
        invite = self._state.room_invites.get(room_id)
        assert invite, Errors.NO_INVITE
        invite.accept()

    def reject_invite(self, room_id: str):
        invite = self._state.room_invites.get(room_id)
        assert invite, Errors.NO_INVITE
        invite.reject()

    async def request_voice(self, room_id: str):
        room = self._state.rooms.get(room_id)
        assert room, Errors.ROOM_NOT_JOINED
        await room.muc_request_voice()



import json
from dataclasses import dataclass, field
from queue import Queue
from threading import Event
from typing import Any, Optional, Union

from websocket import WebSocketApp

from ._common import DEFAULT_TIMEOUT


@dataclass
class MopidyTask:
    """
    A task to be executed by the Mopidy client.
    """

    id: int
    method: str
    args: dict = field(default_factory=dict)
    response: Optional[Any] = None
    response_ready: Event = field(default_factory=Event)
    response_queue: Queue = field(default_factory=Queue)

    def to_dict(self):
        return {
            "jsonrpc": "2.0",
            "id": self.id,
            "method": self.method,
            'params': self.args,
        }

    def __str__(self):
        return json.dumps(self.to_dict())

    def send(self, ws: WebSocketApp):
        assert ws, "Websocket connection not established"
        self.response_ready.clear()
        ws.send(str(self))

    def get_response(self, timeout: Optional[float] = DEFAULT_TIMEOUT) -> Any:
        ret = self.response_queue.get(timeout=timeout)
        if isinstance(ret, dict):
            ret = ret.get('result')

        return ret

    def put_response(self, response: Union[dict, Exception]):
        self.response = response
        self.response_ready.set()
        self.response_queue.put_nowait(response)

    def wait(self, timeout: Optional[float] = DEFAULT_TIMEOUT) -> bool:
        return self.response_ready.wait(timeout=timeout)

    def run(self, ws: WebSocketApp, timeout: Optional[float] = DEFAULT_TIMEOUT):
        self.send(ws)
        return self.get_response(timeout=timeout)



#!/usr/bin/env python

import json
import os

from setuptools import setup, find_packages


def path(fname=''):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), fname))


def readfile(fname):
    with open(path(fname)) as f:
        return f.read()


def pkg_files(dir):
    paths = []
    for p, _, files in os.walk(dir):
        for file in files:
            paths.append(os.path.join('..', p, file))
    return paths


def scan_manifests():
    for root, _, files in os.walk('platypush'):
        for file in files:
            if file == 'manifest.json':
                yield os.path.join(root, file)


def parse_deps(deps):
    ret = []
    for dep in deps:
        if dep.startswith('git+'):
            continue  # Don't include git dependencies in the setup.py, or Twine will complain

        ret.append(dep)

    return ret


def parse_manifest(manifest_file):
    with open(manifest_file) as f:
        manifest = json.load(f).get('manifest')
        if not manifest:
            return None, None

        name = '.'.join(manifest['package'].split('.')[2:])
        return name, parse_deps(manifest.get('install', {}).get('pip', []))


def parse_manifests():
    ret = {}
    for manifest_file in scan_manifests():
        name, deps = parse_manifest(manifest_file)
        if deps:
            ret[name] = deps

    return ret


plugins = pkg_files('platypush/plugins')
backend = pkg_files('platypush/backend')

setup(
    name="platypush",
    version="1.1.1",
    author="Fabio Manganiello",
    author_email="fabio@manganiello.tech",
    description="Platypush service",
    license="MIT",
    python_requires='>= 3.6',
    keywords="home-automation automation iot mqtt websockets redis dashboard notifications",
    url="https://platypush.tech",
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    package_data={
        'platypush': [
            'migrations/alembic.ini',
            'migrations/alembic/*',
            'migrations/alembic/**/*',
            'install/**',
            'install/scripts/*',
            'install/scripts/**/*',
            'install/requirements/*',
            'install/docker/*',
            'components.json.gz',
        ],
    },
    entry_points={
        'console_scripts': [
            'platypush=platypush:main',
            'platydock=platypush.platydock:main',
            'platyvenv=platypush.platyvenv:main',
        ],
    },
    long_description=readfile('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
    ],
    install_requires=[
        'alembic',
        'croniter',
        'docutils',
        'flask',
        'frozendict',
        'marshmallow',
        'marshmallow_dataclass',
        'psutil',
        'python-dateutil',
        'python-magic',
        'pyyaml',
        'redis',
        'requests',
        'rsa',
        'sqlalchemy',
        'tornado',
        'websocket-client',
        'websockets',
        'wheel',
        'zeroconf>=0.27.0',
    ],
    extras_require=parse_manifests(),
)



from abc import ABC, abstractmethod
from logging import getLogger
from typing import List, Optional
from urllib.parse import quote_plus

from .._base import TorrentSearchProvider
from .._model import TorrentSearchResult


class TorrentsCsvBaseProvider(TorrentSearchProvider, ABC):
    """
    Base class for Torrents.csv search providers.
    """

    _http_timeout = 20
    _magnet_trackers = [
        'http://125.227.35.196:6969/announce',
        'http://210.244.71.25:6969/announce',
        'http://210.244.71.26:6969/announce',
        'http://213.159.215.198:6970/announce',
        'http://37.19.5.139:6969/announce',
        'http://37.19.5.155:6881/announce',
        'http://87.248.186.252:8080/announce',
        'http://asmlocator.ru:34000/1hfZS1k4jh/announce',
        'http://bt.evrl.to/announce',
        'http://bt.rutracker.org/ann',
        'https://www.artikelplanet.nl',
        'http://mgtracker.org:6969/announce',
        'http://tracker.baravik.org:6970/announce',
        'http://tracker.dler.org:6969/announce',
        'http://tracker.filetracker.pl:8089/announce',
        'http://tracker.grepler.com:6969/announce',
        'http://tracker.mg64.net:6881/announce',
        'http://tracker.tiny-vps.com:6969/announce',
        'http://tracker.torrentyorg.pl/announce',
        'https://internet.sitelio.me/',
        'https://computer1.sitelio.me/',
        'udp://168.235.67.63:6969',
        'udp://37.19.5.155:2710',
        'udp://46.148.18.250:2710',
        'udp://46.4.109.148:6969',
        'udp://computerbedrijven.bestelinks.nl/',
        'udp://computerbedrijven.startsuper.nl/',
        'udp://computershop.goedbegin.nl/',
        'udp://c3t.org',
        'udp://allerhandelenlaag.nl',
        'udp://tracker.opentrackr.org:1337',
        'udp://tracker.publicbt.com:80',
        'udp://tracker.tiny-vps.com:6969',
        'udp://tracker.openbittorrent.com:80',
        'udp://opentor.org:2710',
        'udp://tracker.ccc.de:80',
        'udp://tracker.blackunicorn.xyz:6969',
        'udp://tracker.coppersurfer.tk:6969',
        'udp://tracker.leechers-paradise.org:6969',
    ]

    def __init__(
        self,
        trackers: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        :param trackers: List of additional trackers to use.
        """
        super().__init__(**kwargs)
        self.logger = getLogger(self.__class__.__name__)
        self.trackers = list({*self._magnet_trackers, *(trackers or [])})

    @classmethod
    def provider_name(cls) -> str:
        return 'torrents.csv'

    @abstractmethod
    def _search(  # pylint: disable=arguments-differ
        self, query: str, *_, limit: int, page: int, **__
    ) -> List[TorrentSearchResult]:
        """
        To be implemented by subclasses.

        :param query: Query string.
        :param limit: Number of results to return (default: 25).
        :param page: Page number (default: 1).
        """

    def _to_magnet(self, info_hash: str, torrent_name: str) -> str:
        """
        Generate a magnet link from an info hash and torrent name.

        :param info_hash: Torrent info hash.
        :param torrent_name: Torrent name.
        :return: Magnet link.
        """
        return (
            f'magnet:?xt=urn:btih:{info_hash}&dn={quote_plus(torrent_name)}&tr='
            + '&tr='.join([quote_plus(tracker) for tracker in self.trackers])
        )


# vim:sw=4:ts=4:et:



command_classes = {
    0x00: 'no_operation',
    0x20: 'basic',
    0x21: 'controller_replication',
    0x22: 'application_status',
    0x23: 'zip_services',
    0x24: 'zip_server',
    0x25: 'switch_binary',
    0x26: 'switch_multilevel',
    0x27: 'switch_all',
    0x28: 'switch_toggle_binary',
    0x29: 'switch_toggle_multilevel',
    0x2a: 'chimney_fan',
    0x2b: 'scene_activation',
    0x2c: 'scene_actuator_conf',
    0x2d: 'scene_controller_conf',
    0x2e: 'zip_client',
    0x2f: 'zip_adv_services',
    0x30: 'sensor_binary',
    0x31: 'sensor_multilevel',
    0x32: 'meter',
    0x33: 'color',
    0x34: 'zip_adv_client',
    0x35: 'meter_pulse',
    0x3c: 'meter_tbl_config',
    0x3d: 'meter_tbl_monitor',
    0x3e: 'meter_tbl_pulse',
    0x38: 'thermostat_heating',
    0x40: 'thermostat_mode',
    0x42: 'thermostat_operating_state',
    0x43: 'thermostat_setpoint',
    0x44: 'thermostat_fan_mode',
    0x45: 'thermostat_fan_state',
    0x46: 'climate_control_schedule',
    0x47: 'thermostat_setback',
    0x4c: 'door_lock_logging',
    0x4e: 'schedule_entry_lock',
    0x50: 'basic_window_covering',
    0x51: 'mtp_window_covering',
    0x56: 'crc16_encap',
    0x5a: 'device_reset_locally',
    0x5b: 'central_scene',
    0x5e: 'zwave_plus_info',
    0x5d: 'antitheft',
    0x60: 'multi_instance',
    0x62: 'door_lock',
    0x63: 'user_code',
    0x66: 'barrier_operator',
    0x70: 'configuration',
    0x71: 'notification',
    0x72: 'manufacturer_specific',
    0x73: 'powerlevel',
    0x75: 'protection',
    0x76: 'lock',
    0x77: 'node_naming',
    0x79: 'sound_switch',
    0x7a: 'firmware_update_md',
    0x7b: 'grouping_name',
    0x7c: 'remote_association_activate',
    0x7d: 'remote_association',
    0x80: 'battery',
    0x81: 'clock',
    0x82: 'hail',
    0x84: 'wake_up',
    0x85: 'association',
    0x86: 'version',
    0x87: 'indicator',
    0x88: 'proprietary',
    0x89: 'language',
    0x8a: 'time',
    0x8b: 'time_parameters',
    0x8c: 'geographic_location',
    0x8d: 'composite',
    0x8e: 'multi_instance_association',
    0x8f: 'multi_cmd',
    0x90: 'energy_production',
    0x91: 'manufacturer_proprietary',
    0x92: 'screen_md',
    0x93: 'screen_attributes',
    0x94: 'simple_av_control',
    0x95: 'av_content_directory_md',
    0x96: 'av_renderer_status',
    0x97: 'av_content_search_md',
    0x98: 'security',
    0x99: 'av_tagging_md',
    0x9a: 'ip_configuration',
    0x9b: 'association_command_configuration',
    0x9c: 'sensor_alarm',
    0x9d: 'silence_alarm',
    0x9e: 'sensor_configuration',
    0xef: 'mark',
    0xf0: 'non_interoperable'
}

command_class_by_name = {
    name: code
    for code, name in command_classes.items()
}



import json
import threading
import time
from typing import Dict, Union

import yaml

from platypush.backend.http.utils import HttpUtils
from platypush.config import Config
from platypush.plugins import Plugin, action
from platypush.procedure import Procedure
from platypush.utils import get_enabled_plugins


class UtilsPlugin(Plugin):
    """
    A plugin for general-purpose util methods
    """

    _DEFAULT_TIMEOUT_PREFIX = '_PlatypushTimeout_'
    _timeout_hndl_idx = 0
    _timeout_hndl_idx_lock = threading.RLock()

    _DEFAULT_INTERVAL_PREFIX = '_PlatypushInterval_'
    _interval_hndl_idx = 0
    _interval_hndl_idx_lock = threading.RLock()

    _pending_timeouts: Dict[str, Union[Procedure, threading.Timer]] = {}
    _pending_intervals: Dict[str, Union[Procedure, threading.Thread]] = {}
    _pending_timeouts_lock = threading.RLock()
    _pending_intervals_lock = threading.RLock()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._plugins = {}
        self._plugins_lock = threading.RLock()

    @action
    def sleep(self, seconds):
        """
        Make the current executor sleep for the specified number of seconds.

        :param seconds: Sleep seconds
        :type seconds: float
        """

        time.sleep(seconds)

    @action
    def set_timeout(self, seconds, actions, name=None, **args):
        """
        Define a set of actions to run after the specified amount of `seconds`.

        :param seconds: Number of seconds before running the timeout procedure
        :type seconds: float

        :param actions: List of actions to be executed after the timeout expires
        :type actions: list[dict]

        :param name: Set an optional name for this timeout. It is advised to set
            a name if you are planning to programmatically cancel the timeout in
            your business logic.
        :type name: str

        :param args: Optional arguments/context to pass to the timeout function
        """

        with self._timeout_hndl_idx_lock:
            self._timeout_hndl_idx += 1
            if not name:
                name = self._DEFAULT_TIMEOUT_PREFIX + str(self._timeout_hndl_idx)
            if name in self._pending_timeouts:
                return (None, f"A timeout named '{name}' is already awaiting")

        procedure = Procedure.build(name=name, requests=actions, _async=False)
        self._pending_timeouts[name] = procedure

        def _proc_wrapper(procedure, **kwargs):
            try:
                procedure.execute(**kwargs)
            finally:
                with self._pending_timeouts_lock:
                    if name in self._pending_timeouts:
                        del self._pending_timeouts[name]

        with self._pending_timeouts_lock:
            self._pending_timeouts[name] = threading.Timer(
                seconds, _proc_wrapper, args=[procedure], kwargs=args
            )
        self._pending_timeouts[name].start()

    @action
    def clear_timeout(self, name):
        """
        Clear a pending timeout procedure

        :param name: Name of the timeout to clear
        :type name: str
        """
        with self._pending_timeouts_lock:
            if name not in self._pending_timeouts:
                self.logger.debug('%s is not a pending timeout', name)
                return
            timer = self._pending_timeouts.pop(name)

        timer.cancel()

    @action
    def get_timeouts(self):
        """
        Get info about the pending timeouts

        :returns: dict.

        Example::

            {
                "test_timeout": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        """

        response = {}

        for name in self._pending_timeouts:
            # pylint: disable=no-member
            response[name] = self.get_timeout(name).output.get(name)
        return response

    @action
    def get_timeout(self, name):
        """
        Get info about a pending timeout

        :param name: Name of the timeout to get
        :type name: str

        :returns: dict

        Example::

            {
                "test_timeout": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        If no such timeout exist with the specified name then the value of the
        timeout name will be null.
        """

        response = {name: None}

        with self._pending_timeouts_lock:
            timer = self._pending_timeouts.get(name)
            if not timer:
                return response

            return {
                name: {
                    'seconds': timer.interval,
                    'actions': [json.loads(str(a)) for a in timer.args[0].requests],
                }
            }

    @action
    def set_interval(self, seconds, actions, name=None, **args):
        """
        Define a set of actions to run each specified amount of `seconds`.

        :param seconds: Number of seconds between two runs of the interval
            procedure
        :type seconds: float

        :param actions: List of actions to be executed at each interval
        :type actions: list[dict]

        :param name: Set an optional name for this interval. It is advised to
            set a name if you are planning to programmatically cancel the
            interval in your business logic.
        :type name: str

        :param args: Optional arguments/context to pass to the interval function
        """

        with self._interval_hndl_idx_lock:
            self._interval_hndl_idx += 1
            if not name:
                name = self._DEFAULT_INTERVAL_PREFIX + str(self._interval_hndl_idx)

            if name in self._pending_intervals:
                return (None, f"An interval named '{name}' is already running")

        procedure = Procedure.build(name=name, requests=actions, _async=False)
        self._pending_intervals[name] = procedure

        def _proc_wrapper(procedure, seconds, **kwargs):
            while True:
                with self._pending_intervals_lock:
                    if name not in self._pending_intervals:
                        return

                procedure.execute(**kwargs)
                time.sleep(seconds)

        with self._pending_intervals_lock:
            self._pending_intervals[name] = threading.Thread(
                target=_proc_wrapper, args=[procedure, seconds], kwargs=args
            )
        self._pending_intervals[name].start()

    @action
    def clear_interval(self, name):
        """
        Clear a running interval procedure

        :param name: Name of the interval to clear
        :type name: str
        """
        with self._pending_intervals_lock:
            if name not in self._pending_intervals:
                self.logger.debug('%s is not a running interval', name)
                return
            del self._pending_intervals[name]

    @action
    def get_intervals(self):
        """
        Get info about the running intervals

        :returns: dict

        Example::

            {
                "test_interval": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        """

        response = {}

        for name in self._pending_intervals:
            # pylint: disable=no-member
            response[name] = self.get_interval(name).output.get(name)
        return response

    @action
    def get_interval(self, name):
        """
        Get info about a running interval

        :param name: Name of the interval to get
        :type name: str

        :returns: dict. Example:

        .. code-block:: json

            {
                "test_interval": {
                    "seconds": 10.0,
                    "actions": [
                        {
                            "action": "action_1",
                            "args": {
                                "name_1": "value_1"
                            }
                        }
                    ]
                }
            }

        If no such interval exist with the specified name then the value of the
        timeout name will be null.
        """

        response = {name: None}

        with self._pending_intervals_lock:
            timer = self._pending_intervals.get(name)
            if not timer:
                return response

            return {
                name: {
                    'seconds': timer._args[1],  # pylint: disable=protected-access
                    'actions': [
                        json.loads(str(a))
                        for a in (
                            # pylint: disable=protected-access
                            timer._args[0].requests
                        )
                    ],
                }
            }

    @action
    def search_directory(self, directory, extensions, recursive=False):
        return HttpUtils.search_directory(directory, recursive=recursive, *extensions)

    @action
    def search_web_directory(self, directory, extensions):
        return HttpUtils.search_web_directory(directory, *extensions)

    @action
    def get_enabled_plugins(self) -> dict:
        """
        :return: The list of enabled plugins as a ``name -> configuration`` map.
        """
        if self._plugins:
            return self._plugins

        plugins = {}
        with self._plugins_lock:
            for name in get_enabled_plugins():
                plugins[name] = Config.get(name)

        return plugins

    @action
    def rst_to_html(self, text: str):
        """
        Utility action to convert RST to HTML.

        It is mostly used by the frontend to render the docstring of the
        available plugins and actions.
        """
        try:
            import docutils.core  # type: ignore
        except ImportError:
            self.logger.warning(
                "docutils is not installed. "
                "Please install docutils to convert RST to HTML."
            )
            return text

        return docutils.core.publish_parts(text, writer_name='html')['html_body']

    @action
    def to_yaml(self, obj: Union[dict, list, tuple, str, int, float, bool]):
        """
        Convert an object to YAML
        """
        return yaml.dump(obj, indent=2)


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import datetime
from logging import getLogger
import os
import queue
from threading import Event, RLock, Thread
import time
from typing import IO, Callable, Final, Generator, Optional, Tuple, Type, Union

import sounddevice as sd

from platypush.context import get_bus
from platypush.message.event.sound import SoundEvent
from platypush.utils import get_redis

from .._converters import AudioConverter
from .._model import AudioState, StreamType

_StreamType = Union[sd.Stream, sd.OutputStream]


class AudioThread(Thread, ABC):
    """
    Base class for audio play/record stream threads.
    """

    _DEFAULT_FILE: Final[str] = os.devnull
    """Unless otherwise specified, the audio streams will be sent to /dev/null"""
    _DEFAULT_CONVERTER_TIMEOUT: Final[float] = 1

    def __init__(
        self,
        device: Union[str, Tuple[str, str]],
        channels: int,
        volume: float,
        sample_rate: int,
        dtype: str,
        blocksize: int,
        ffmpeg_bin: str = 'ffmpeg',
        stream: bool = False,
        audio_pass_through: bool = False,
        infile: Optional[str] = None,
        outfile: Optional[str] = None,
        duration: Optional[float] = None,
        input_format: Optional[str] = None,
        output_format: Optional[str] = None,
        latency: Union[float, str] = 'high',
        redis_queue: Optional[str] = None,
        should_stop: Optional[Event] = None,
        converter_timeout: Optional[float] = None,
        stream_name: Optional[str] = None,
        queue_size: Optional[int] = None,
        **kwargs,
    ):
        """
        :param device: Audio device to use.
        :param channels: Number of channels to use.
        :param volume: Input/output volume, between 0 and 100.
        :param sample_rate: Sample rate to use.
        :param dtype: Data type to use.
        :param blocksize: Block size to use.
        :param ffmpeg_bin: Path to the ffmpeg binary.
        :param stream: Whether to stream the audio to Redis consumers.
        :param audio_pass_through: Whether to pass the audio through to the
            application's output stream.
        :param infile: Path to the input file or URL, if this is an output
            stream.
        :param outfile: Path to the output file.
        :param duration: Duration of the audio stream.
        :param input_format: Input format override.
        :param output_format: Output format override.
        :param latency: Latency to use.
        :param redis_queue: Redis queue to use.
        :param should_stop: Synchronize with upstream stop events.
        :param converter_timeout: How long to wait for the converter to finish.
        :param stream_name: Name of the stream.
        :param queue_size: Maximum size of the audio queue.
        """
        super().__init__(**kwargs)

        self.device = device
        self.outfile = os.path.expanduser(outfile or self._DEFAULT_FILE)
        self.infile = os.path.expanduser(infile or self._DEFAULT_FILE)
        self.ffmpeg_bin = ffmpeg_bin
        self.channels = channels
        self.volume = volume
        self.sample_rate = sample_rate
        self.dtype = dtype
        self.input_format = input_format
        self.output_format = output_format
        self.stream = stream
        self.duration = duration
        self.blocksize = blocksize * channels
        self.latency = latency
        self._redis_queue = redis_queue
        self.audio_pass_through = audio_pass_through
        self.queue_size = queue_size
        self._stream_name = stream_name
        self.logger = getLogger(__name__)

        self._state = AudioState.STOPPED
        self._state_lock = RLock()
        self._started_time: Optional[float] = None
        self._converter: Optional[AudioConverter] = None
        self._should_stop = should_stop or Event()
        self._converter_timeout = converter_timeout or self._DEFAULT_CONVERTER_TIMEOUT
        self.audio_stream: Optional[_StreamType] = None
        self.stream_index: Optional[int] = None
        self.paused_changed = Event()
        self._converter_terminated = Event()

    @property
    def should_stop(self) -> bool:
        """
        Proxy for `._should_stop.is_set()`.
        """
        return self._should_stop.is_set() or bool(
            self.state == AudioState.STOPPED and self._started_time
        )

    @property
    def gain(self) -> float:
        return self.volume / 100

    def wait_stop(self, timeout: Optional[float] = None):
        """
        Wait for the stop signal to be received.
        """
        return self._should_stop.wait(timeout=timeout)

    def _audio_callback(self) -> Callable:
        """
        Returns a callback to handle the raw frames captures from the audio device.
        """

        def empty_callback(*_, **__):
            pass

        return empty_callback

    @property
    def stream_name(self) -> str:
        if self._stream_name:
            return self._stream_name

        ret = f'platypush:audio:{self.direction.value}'
        if self.stream_index is not None:
            ret += f':{self.stream_index}'
        return ret

    @stream_name.setter
    def stream_name(self, value: Optional[str]):
        self._stream_name = value

    @property
    @abstractmethod
    def direction(self) -> StreamType:
        """
        The default direction for this stream - input or output.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def _audio_converter_type(self) -> Optional[Type[AudioConverter]]:
        """
        This property indicates the type that should be used for the audio
        converter.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def _started_event_type(self) -> Type[SoundEvent]:
        """
        Event type that will be emitted when the audio starts.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def _stopped_event_type(self) -> Type[SoundEvent]:
        """
        Event type that will be emitted when the audio stops.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def _paused_event_type(self) -> Type[SoundEvent]:
        """
        Event type that will be emitted when the audio is paused.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def _resumed_event_type(self) -> Type[SoundEvent]:
        """
        Event type that will be emitted when the audio is resumed.
        """
        raise NotImplementedError()

    @property
    def _stream_type(self) -> Union[Type[sd.Stream], Type[sd.OutputStream]]:
        """
        The type of stream this thread is mapped to.
        """
        return sd.Stream

    @property
    def _converter_args(self) -> dict:
        """
        Extra arguments to pass to the audio converter.
        """
        return {}

    @property
    def _stream_args(self) -> dict:
        """
        Extra arguments to pass to the stream constructor.
        """
        return {}

    @property
    def redis_queue(self) -> str:
        """
        Redis queue for audio streaming.
        """
        if self._redis_queue:
            return self._redis_queue

        dev = (
            self.device
            if isinstance(self.device, (str, int))
            else '-'.join(map(str, self.device))
        )

        name = f'platypush-audio-stream-{self.__class__.__name__}-{dev}'
        if self.stream_index is not None:
            name = f'{name}-{self.stream_index}'

        return name

    def _on_audio_converted(self, data: bytes, out_f: Optional[IO] = None):
        """
        This callback will be called when the audio data has been converted.
        """
        if out_f:
            out_f.write(data)

        if self.stream:
            get_redis().publish(self.redis_queue, data)

    def _wait_running(self):
        """
        If the stream is in paused state, wait for the state to change.
        """
        while self.state == AudioState.PAUSED:
            self.paused_changed.wait()

    def main(
        self,
        converter: Optional[AudioConverter] = None,
        out_f: Optional[IO] = None,
    ):
        """
        Main loop.
        """
        self.notify_start()

        self.logger.info(
            'Started %s on device [%s]', self.__class__.__name__, self.device
        )
        self._started_time = time.time()

        while not self.should_stop and (
            self.duration is None or time.time() - self._started_time < self.duration
        ):
            self._wait_running()
            if not converter:
                self.wait_stop(0.1)
                continue

            if self.should_stop:
                break

            timeout = (
                max(
                    0,
                    min(
                        self.duration - (time.time() - self._started_time),
                        self._converter_timeout,
                    ),
                )
                if self.duration is not None
                else self._converter_timeout
            )

            should_continue = self._process_converted_audio(
                converter, timeout=timeout, out_f=out_f
            )

            if not should_continue:
                break

    def _process_converted_audio(
        self, converter: AudioConverter, timeout: float, out_f: Optional[IO]
    ) -> bool:
        """
        It reads the converted audio from the converter and passes it downstream.

        :return: True if the process should continue, False if it should terminate.
        """
        data = converter.read(timeout=timeout)
        if not data:
            return self._on_converter_timeout(converter)

        self._on_audio_converted(data, out_f)
        return True

    def _on_converter_timeout(self, converter: AudioConverter) -> bool:
        """
        Callback logic invoked if the converter times out.

        :return: ``True`` (default) if the thread is supposed to continue,
            ``False`` if it should terminate.
        """
        self.logger.debug('Timeout on converter %s', converter.__class__.__name__)
        # Continue only if the converter hasn't terminated
        return not self._converter_terminated.is_set()

    def run(self):
        """
        Wrapper for the main loop that initializes the converter and the stream.
        """
        super().run()
        self.paused_changed.clear()

        try:
            with self.open_converter() as converter, self._stream_type(
                samplerate=self.sample_rate,
                device=self.device,
                channels=self.channels,
                dtype=self.dtype,
                latency=self.latency,
                blocksize=self.blocksize,
                **self._stream_args,
            ) as self.audio_stream, open(
                self.outfile, 'wb'
            ) as out_f, self._audio_generator():
                self.main(converter=converter, out_f=out_f)
        except queue.Empty:
            self.logger.warning(
                'Audio callback timeout for %s', self.__class__.__name__
            )
        except Exception as e:
            self.logger.warning('Unhandled sound on %s', self.__class__.__name__)
            self.logger.exception(e)
        finally:
            self.notify_stop()

    @contextmanager
    def _audio_generator(self) -> Generator[Optional[Thread], None, None]:
        """
        :yield: A <Thread, Queue> pair where the thread generates raw audio
            frames (as numpy arrays) that are sent to the specified queue.
        """
        yield None

    @contextmanager
    def open_converter(self) -> Generator[Optional[AudioConverter], None, None]:
        """
        Context manager for the converter process.
        """
        if self._audio_converter_type is None:
            yield None
            return

        assert not self._converter, 'A converter process is already running'
        self._converter = self._audio_converter_type(
            ffmpeg_bin=self.ffmpeg_bin,
            sample_rate=self.sample_rate,
            channels=self.channels,
            volume=self.volume,
            dtype=self.dtype,
            chunk_size=self.blocksize,
            on_exit=self._converter_terminated.set,
            **self._converter_args,
        )

        self._converter.start()
        yield self._converter

        self._converter.stop()
        self._converter.join(timeout=2)
        self._converter = None

    @contextmanager
    def _change_state(self, state: AudioState, event_type: Type[SoundEvent]):
        """
        Changes the state and it emits the specified event if the state has
        actually changed.

        It uses a context manager pattern, and everything in between will be
        executed before the events are dispatched.
        """
        with self._state_lock:
            prev_state = self.state
            self.state = state

        yield
        if prev_state != state:
            self._notify(event_type)

    def _notify(self, event_type: Type[SoundEvent], **kwargs):
        """
        Notifies the specified event.
        """
        get_bus().post(event_type(device=self.device, **kwargs))

    def notify_start(self):
        """
        Notifies the start event.
        """
        with self._change_state(AudioState.RUNNING, self._started_event_type):
            pass

    def notify_stop(self):
        """
        Notifies the stop event.
        """
        with self._change_state(AudioState.STOPPED, self._stopped_event_type):
            if self._converter:
                self._converter.stop()
                self.paused_changed.set()
                self.paused_changed.clear()

    def notify_pause(self):
        """
        Notifies a pause toggle event.
        """
        states = {
            AudioState.PAUSED: AudioState.RUNNING,
            AudioState.RUNNING: AudioState.PAUSED,
        }

        with self._state_lock:
            new_state = states.get(self.state)
            if not new_state:
                return

            event_type = (
                self._paused_event_type
                if new_state == AudioState.PAUSED
                else self._resumed_event_type
            )

            with self._change_state(new_state, event_type):
                self.paused_changed.set()
                self.paused_changed.clear()

    @property
    def state(self):
        """
        Thread-safe wrapper for the stream state.
        """
        with self._state_lock:
            return self._state

    @state.setter
    def state(self, value: AudioState):
        """
        Thread-safe setter for the stream state.
        """
        with self._state_lock:
            self._state = value

    def asdict(self) -> dict:
        """
        Serialize the thread information.
        """
        return {
            'device': self.device,
            'outfile': self.outfile,
            'infile': self.infile,
            'direction': self.direction,
            'ffmpeg_bin': self.ffmpeg_bin,
            'channels': self.channels,
            'sample_rate': self.sample_rate,
            'dtype': self.dtype,
            'streaming': self.stream,
            'duration': self.duration,
            'blocksize': self.blocksize,
            'latency': self.latency,
            'redis_queue': self.redis_queue,
            'audio_pass_through': self.audio_pass_through,
            'state': self._state.value,
            'volume': self.volume,
            'started_time': datetime.fromtimestamp(self._started_time)
            if self._started_time
            else None,
            'stream_index': self.stream_index,
            'stream_name': self.stream_name,
        }


# vim:sw=4:ts=4:et:



import os
import pathlib
import time

from collections import defaultdict
from threading import Thread, RLock
from typing import Any, Dict, List, Optional, Union, Collection

from platypush.config import Config
from platypush.message.event.mail import (
    FlaggedMailEvent,
    SeenMailEvent,
    UnflaggedMailEvent,
    UnseenMailEvent,
)
from platypush.plugins import RunnablePlugin, action

from ._account import Account
from ._model import FolderStatus, Mail, MailFlagType, AccountsStatus
from ._plugin import MailInPlugin, MailOutPlugin

AccountType = Union[str, int, Account]
AccountFolderChanges = Dict[str, Dict[MailFlagType, Dict[int, bool]]]


class MailPlugin(RunnablePlugin):
    """
    Plugin to:

       - Monitor one or more mailboxes, and emit events when new messages are
         received, seen, flagged or unflagged.

       - Send mail messages.

       - Search for messages in a mailbox.

    """

    def __init__(
        self,
        accounts: List[Dict[str, Any]],
        timeout: float = 20.0,
        poll_interval: float = 60.0,
        **kwargs,
    ):
        """
        Example accounts configuration:

          .. code-block:: yaml

            mail:
                # Display name to be used for outgoing emails. Default:
                # the `from` parameter will be used from :meth:`.send`,
                # and, if missing, the username from the account configuration
                # will be used.
                display_name: My Name

                # How often we should poll for updates (default: 60 seconds)
                poll_interval: 60

                # Connection timeout (default: 20 seconds)
                # Can be overridden on a per-account basis
                timeout: 20

                # Domain to be used for outgoing emails. Default: inferred
                # from the account configuration
                domain: example.com

                accounts:
                    - name: "My Local Account"
                      username: me@mydomain.com
                      password: my-password

                      # The default flag sets this account as the default one
                      # for mail retrieval and sending if no account is
                      # specified on an action. If multiple accounts are set
                      # and none is set as default, and no account is specified
                      # on an action, then the first configured account will be
                      # used.
                      default: true

                      # Alternatively, you can run an external command
                      # to get the password
                      # password_cmd: "pass show mail/example.com"

                      # Path to a custom certfile if the mail server uses a
                      # self-signed certificate
                      # certfile: /path/to/certfile

                      # Path to a custom keyfile if the mail server requires
                      # client authentication. It requires certfile to be set
                      # too
                      # keyfile: /path/to/keyfile

                      incoming:
                          # Supported protocols: imap, imaps
                          server: imaps://mail.example.com:993

                      outgoing:
                          # The `incoming` and `outgoing` configurations can
                          # override the global `username` and `password` and
                          # other authentication parameters of the account
                          username: me
                          password: my-smtp-password

                          # Supported protocols: smtp, smtps, smtp+starttls,
                          server: smtps://mail.example.com:465

                      # These folders will be monitored for new messages
                      monitor_folders:
                          - All Mail

                    - name: "GMail"
                      username: me@gmail.com

                      # Access token, if the mail server supports OAuth2
                      # access_token: my-access-token

                      # OAuth mechanism, if the mail server supports OAuth2
                      # (default: XOAUTH2)
                      # oauth_mechanism: XOAUTH2

                      # OAuth vendor, if the mail server supports OAuth2
                      # oauth_vendor: GOOGLE

                      incoming:
                          # Defaults to port 993 for IMAPS if no port is
                          # specified on the URL
                          server: imaps://imap.gmail.com

                      outgoing:
                          # Defaults to port 465 for SMTPS if no port is
                          # specified on the URL
                          server: smtps://smtp.gmail.com

                      monitor_folders:
                          - INBOX
                          - Sent

        :param accounts: List of available mailboxes/accounts.
        :param poll_interval: How often the plugin should poll for new messages
            (default: 60 seconds).
        :param timeout: Timeout for the mail server connection (default: 20
            seconds).
        """
        assert accounts, 'No mail accounts configured'

        super().__init__(poll_interval=poll_interval, **kwargs)
        self.timeout = timeout
        self.accounts = self._parse_accounts(accounts)
        self._accounts_by_name = {acc.name: acc for acc in self.accounts}
        self._default_account = next(
            (acc for acc in self.accounts if acc.default), self.accounts[0]
        )

        self._status = AccountsStatus()
        self._db_lock = RLock()
        self.workdir = os.path.join(Config.get_workdir(), 'mail')
        self._status_file = os.path.join(self.workdir, 'status.json')

        # Configure/sync db
        pathlib.Path(self.workdir).mkdir(parents=True, exist_ok=True, mode=0o750)
        self._load_status()

    def _load_status(self):
        with self._db_lock:
            try:
                with open(self._status_file) as f:
                    self._status = AccountsStatus.read(f)
            except FileNotFoundError:
                self._status = AccountsStatus()
            except Exception as e:
                self.logger.warning(
                    'Could not load mail status from %s: %s', self._status_file, e
                )
                self._status = AccountsStatus()

    def _get_account(self, account: Optional[AccountType] = None) -> Account:
        if isinstance(account, Account):
            return account

        if isinstance(account, int):
            account -= 1
            assert (
                0 <= account < len(self.accounts)
            ), f'Invalid account index {account} (valid range: 1-{len(self.accounts)})'

            return self.accounts[account]

        if isinstance(account, str):
            acc = self._accounts_by_name.get(account)
            assert acc, f'No account found with name "{account}"'
            return acc

        return self._default_account

    def _get_in_plugin(self, account: Optional[AccountType] = None) -> MailInPlugin:
        acc = self._get_account(account)
        assert acc.incoming, f'No incoming configuration found for account "{acc.name}"'
        return acc.incoming

    def _get_out_plugin(self, account: Optional[AccountType] = None) -> MailOutPlugin:
        acc = self._get_account(account)
        assert acc.outgoing, f'No outgoing configuration found for account "{acc.name}"'
        return acc.outgoing

    def _parse_accounts(self, accounts: List[Dict[str, Any]]) -> List[Account]:
        ret = []
        for i, acc in enumerate(accounts):
            idx = i + 1
            name = acc.pop('name') if 'name' in acc else f'Account #{idx}'
            incoming_conf = acc.pop('incoming')
            outgoing_conf = acc.pop('outgoing')
            monitor_folders = acc.pop('monitor_folders', [])

            assert (
                incoming_conf or outgoing_conf
            ), f'No incoming/outgoing configuration specified for account "{name}"'

            if monitor_folders:
                assert incoming_conf, (
                    f'Cannot monitor folders for account "{name}" '
                    'without incoming configuration'
                )

            acc['poll_interval'] = self.poll_interval
            acc['timeout'] = acc.get('timeout', self.timeout)
            ret.append(
                Account.build(
                    name=name,
                    incoming=incoming_conf,
                    outgoing=outgoing_conf,
                    monitor_folders=monitor_folders,
                    **acc,
                )
            )

        return ret

    @property
    def _monitored_accounts(self):
        return [acc for acc in self.accounts if acc.monitor_folders]

    @property
    def _account_by_name(self) -> Dict[str, Account]:
        return {acc.name: acc for acc in self.accounts}

    @staticmethod
    def _check_thread(plugin: MailInPlugin, folder: str, results: FolderStatus):
        results[MailFlagType.UNREAD] = {
            msg.id: msg for msg in plugin.search_unseen_messages(folder=folder)
        }

        results[MailFlagType.FLAGGED] = {
            msg.id: msg for msg in plugin.search_flagged_messages(folder=folder)
        }

    def _check_mailboxes(self) -> AccountsStatus:
        # Workers indexed by (account_name, folder) -> thread
        workers = {}
        status = AccountsStatus()

        for account in self._monitored_accounts:
            for folder in account.monitor_folders or []:
                worker = Thread(
                    target=self._check_thread,
                    name=f'check-mailbox-{account.name}-{folder}',
                    kwargs={
                        'plugin': account.incoming,
                        'results': status[account.name][folder],
                        'folder': folder,
                    },
                )

                worker.start()
                workers[account.name, folder] = worker

        wait_start = time.time()
        for worker_key, worker in workers.items():
            account = self._account_by_name[worker_key[0]]
            if not account.incoming:
                continue

            # The timeout should be the time elapsed since wait_start + the configured timeout
            timeout = max(
                0, account.incoming.server.timeout - (time.time() - wait_start)
            )
            worker.join(timeout=timeout)
            if worker.is_alive():
                self.logger.warning('Timeout while polling account %s', account.name)

        return status

    @action
    def get_folders(
        self,
        folder: str = '',
        pattern: str = '*',
        account: Optional[AccountType] = None,
    ) -> List[Dict[str, str]]:
        r"""
        Get the list of all the folders hosted on the server or those matching a pattern.

        :param folder: Base folder (default: root).
        :param pattern: Pattern to search (default: None).
        :param account: Account name or index (default: default account).
        :return: Example:

            .. code-block:: json

                [
                    {
                        "name": "INBOX",
                        "flags": "\\Noinferiors",
                        "delimiter": "/"
                    },
                    {
                        "name": "Archive",
                        "flags": "\\Noinferiors",
                        "delimiter": "/"
                    },
                    {
                        "name": "Spam",
                        "flags": "\\Noinferiors",
                        "delimiter": "/"
                    }
                ]

        """
        return self._get_in_plugin(account).get_folders(folder=folder, pattern=pattern)

    @action
    def get_sub_folders(
        self,
        folder: str = '',
        pattern: str = '*',
        account: Optional[AccountType] = None,
    ) -> List[Dict[str, str]]:
        r"""
        Get the list of all the sub-folders hosted on the server or those matching a pattern.

        :param folder: Base folder (default: root).
        :param pattern: Pattern to search (default: None).
        :param account: Account name or index (default: default account).
        :return: Example:

            .. code-block:: json

                [
                    {
                        "name": "INBOX",
                        "flags": "\\Noinferiors",
                        "delimiter": "/"
                    },
                    {
                        "name": "Archive",
                        "flags": "\\Noinferiors",
                        "delimiter": "/"
                    },
                    {
                        "name": "Spam",
                        "flags": "\\Noinferiors",
                        "delimiter": "/"
                    }
                ]

        """
        return self._get_in_plugin(account).get_sub_folders(
            folder=folder, pattern=pattern
        )

    @action
    def search(
        self,
        criteria: Union[str, List[str]] = 'ALL',
        folder: str = 'INBOX',
        attributes: Optional[List[str]] = None,
        account: Optional[AccountType] = None,
    ) -> List[Mail]:
        """
        Search for messages on the server that fit the specified criteria.

        If no criteria is specified, then all the messages in the folder will
        be returned.

        :param criteria: It should be a sequence of one or more criteria items.
            Each criterion item may be either unicode or bytes (default:
            ``ALL``). Example values::

                ['UNSEEN']
                ['FROM', 'me@example.com']
                ['TO', 'me@example.com']
                ['SMALLER', 500]
                ['NOT', 'DELETED']
                ['TEXT', 'foo bar', 'FLAGGED', 'SUBJECT', 'baz']
                ['SINCE', '2020-03-14T12:13:45+00:00']

            It is also possible (but not recommended) to pass the combined
            criteria as a single string. In this case IMAPClient won't perform
            quoting, allowing lower-level specification of criteria. Examples
            of this style::

                'UNSEEN'
                'SMALLER 500'
                'NOT DELETED'
                'TEXT "foo bar" FLAGGED SUBJECT "baz"'
                'SINCE 03-Apr-2005'

            To support complex search expressions, criteria lists can be
            nested. The following will match messages that are both not flagged
            and do not have "foo" in the subject::

                ['NOT', ['SUBJECT', 'foo', 'FLAGGED']]

            See :rfc:`3501#section-6.4.4` for more details.

        :param folder: Folder to search (default: ``INBOX``).
        :param attributes: Attributes that should be retrieved, according to
            `RFC 3501 <https://tools.ietf.org/html/rfc3501>`_
            (default: ``ALL`` = ``[FLAGS INTERNALDATE RFC822.SIZE ENVELOPE]``).
            Note that ``BODY`` will be ignored if specified here for
            performance reasons - use :meth:`.get_message` if you want to get
            the full content of a message known its ID from :meth:`.search`.

        :param account: Account name or index (default: default account).
        :return: List of messages matching the criteria. Example:

        .. code-block:: json

            [
              {
                "id": 702,
                "seq": 671,
                "flags": [
                  "nonjunk"
                ],
                "internal_date": "2020-08-30T00:31:52+00:00",
                "size": 2908738,
                "bcc": {},
                "cc": {},
                "date": "2020-08-30T00:31:52+00:00",
                "from": {
                  "test123@gmail.com": {
                    "name": "A test",
                    "route": null,
                    "email": "test123@gmail.com"
                  }
                },
                "message_id": "<SOMETHING@mail.gmail.com>",
                "in_reply_to": "<SOMETHING@mail.gmail.com>",
                "reply_to": {},
                "sender": {
                  "test123@gmail.com": {
                    "name": "A test",
                    "route": null,
                    "email": "test123@gmail.com"
                  }
                },
                "subject": "Test email",
                "to": {
                  "me@gmail.com": {
                    "name": null,
                    "route": null,
                    "email": "me@gmail.com"
                  }
                }
              }
            ]

        """
        return self._get_in_plugin(account).search(
            criteria=criteria, folder=folder, attributes=attributes
        )

    @action
    def search_unseen_messages(
        self, folder: str = 'INBOX', account: Optional[AccountType] = None
    ) -> List[Mail]:
        """
        Shortcut for :meth:`.search` that returns only the unread messages.
        """
        return self._get_in_plugin(account).search_unseen_messages(folder=folder)

    @action
    def search_flagged_messages(
        self, folder: str = 'INBOX', account: Optional[AccountType] = None
    ) -> List[Mail]:
        """
        Shortcut for :meth:`.search` that returns only the flagged/starred messages.
        """
        return self._get_in_plugin(account).search_flagged_messages(folder=folder)

    @action
    def search_starred_messages(
        self, folder: str = 'INBOX', account: Optional[AccountType] = None
    ) -> List[Mail]:
        """
        Shortcut for :meth:`.search` that returns only the starred messages.
        """
        return self._get_in_plugin(account).search_starred_messages(folder=folder)

    @action
    def sort(
        self,
        folder: str = 'INBOX',
        sort_criteria: Union[str, List[str]] = 'ARRIVAL',
        criteria: Union[str, List[str]] = 'ALL',
        account: Optional[AccountType] = None,
    ) -> List[int]:
        """
        Return a list of message ids from the currently selected folder, sorted
        by ``sort_criteria`` and optionally filtered by ``criteria``. Note that
        SORT is an extension to the IMAP4 standard, so it may not be supported
        by all IMAP servers.

        :param folder: Folder to be searched (default: ``INBOX``).
        :param sort_criteria: It may be a sequence of strings or a single
            string. IMAPClient will take care any required conversions. Valid
            *sort_criteria* values::

              .. code-block:: python

                ['ARRIVAL']
                ['SUBJECT', 'ARRIVAL']
                'ARRIVAL'
                'REVERSE SIZE'

        :param criteria: Optional filter for the messages, as specified in
            :meth:`.search`.
        :param account: Account name or index (default: default account).
        :return: A list of message IDs that fit the criteria.
        """
        return self._get_in_plugin(account).sort(
            folder=folder, sort_criteria=sort_criteria, criteria=criteria
        )

    @action
    def get_message(
        self,
        id: int,  # pylint: disable=redefined-builtin
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
        with_body: bool = True,
    ) -> Mail:
        r"""
        Get the full content of a message given the ID returned by :meth:`.search`.

        :param id: Message ID.
        :param folder: Folder name (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        :param with_body: If set then the body/payload will be included in the response (default: ``True``).
        :return: A message in the same format as :meth:`.search`, with an added
            ``payload`` attribute containing the body/payload. Example response:

            .. code-block:: json

                {
                  "id": 123,
                  "date": "2024-01-13T21:04:50",
                  "size": 3833,
                  "from": {
                    "you@example.com": {
                      "name": "Me",
                      "route": null,
                      "email": "you@example.com"
                    }
                  },
                  "to": {
                    "me@example.com": {
                      "name": "Me",
                      "route": null,
                      "email": "me@example.com"
                    }
                  },
                  "cc": {
                    "they@example.com": {
                      "name": "They",
                      "route": null,
                      "email": "they@example.com"
                    }
                  },
                  "bcc": {
                    "boss@example.com": {
                      "name": "Boss",
                      "route": null,
                      "email": "boss@example.com"
                    }
                  },
                  "subject": "Test email",
                  "content": {
                    "headers": {
                      "Return-Path": "<you@example.com>",
                      "Delivered-To": "me@example.com",
                      "Date": "Sat, 13 Jan 2024 20:04:50 +0000",
                      "To": "Me <me@example.com>",
                      "Cc": "They <they@example.com>",
                      "Bcc": "Boss <boss@example.com",
                      "From": "You <you@example.com>",
                      "Subject": "Test email",
                      "Message-ID": "<0123456789@client>",
                      "MIME-Version": "1.0",
                      "Content-Type": "multipart/mixed;\r\n boundary=\"0123456789\""
                    },
                    "body": "This is the email body",
                    "attachments": [
                      {
                        "filename": "signature.asc",
                        "headers": {
                          "Content-Type": "application/pgp-signature; name=signature.asc",
                          "Content-Transfer-Encoding": "base64",
                          "Content-Disposition": "attachment; filename=signature.asc"
                        },
                        "body": "-----BEGIN PGP SIGNATURE-----\r\n\r\n....\r\n\r\n-----END PGP SIGNATURE-----\r\n"
                      },
                      {
                        "filename": "image.jpg",
                        "body": "/9j/4gIcSUNDX1BST0ZJTEUAA...",
                        "headers": {
                          "Content-Type": "image/jpeg; Name=\"image.jpg\"",
                          "MIME-Version": "1.0",
                          "Content-Transfer-Encoding": "base64",
                          "Content-Disposition": "attachment; filename=\"profile_pic.jpg\""
                        }
                      }
                    ]
                  },
                  "seq": 123,
                  "internal_date": "2024-01-13T21:05:12",
                  "message_id": "<0123456789@client>",
                  "reply_to": {
                    "you@example.com": {
                      "name": "You",
                      "route": null,
                      "email": "you@example.com"
                    }
                  },
                  "sender": {
                    "you@example.com": {
                      "name": "You",
                      "route": null,
                      "email": "you@example.com"
                    }
                  }
                }

        """
        return self._get_in_plugin(account).get_message(
            id=id, folder=folder, with_body=with_body
        )

    @action
    def get_messages(
        self,
        ids: Collection[int],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
        with_body: bool = True,
    ) -> Dict[int, Mail]:
        """
        Get the full content of a list of messages given their IDs returned by :meth:`.search`.

        :param ids: IDs of the messages to retrieve.
        :param folder: Folder name (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        :param with_body: If set then the body/payload will be included in the response
            (default: ``True``).
        :return: A dictionary in the format ``{id -> msg}``, where ``msg`` is in the same
            format as :meth:`.search`, with an added ``payload`` attribute containing the
            body/payload. See :meth:`.get_message` for an example message format.
        """
        return self._get_in_plugin(account).get_messages(
            *ids, folder=folder, with_body=with_body
        )

    @action
    def create_folder(self, folder: str, account: Optional[AccountType] = None):
        """
        Create a folder on the server.

        :param folder: Folder name.
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).create_folder(folder=folder)

    @action
    def rename_folder(
        self, old_name: str, new_name: str, account: Optional[AccountType] = None
    ):
        """
        Rename a folder on the server.

        :param old_name: Previous name
        :param new_name: New name
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).rename_folder(
            old_name=old_name, new_name=new_name
        )

    @action
    def delete_folder(self, folder: str, account: Optional[AccountType] = None):
        """
        Delete a folder from the server.

        :param folder: Folder name.
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).delete_folder(folder=folder)

    @action
    def add_flags(
        self,
        messages: List[int],
        flags: Union[str, List[str]],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Add a set of flags to the specified set of message IDs.

        :param messages: List of message IDs.
        :param flags: List of flags to be added. Examples:

          .. code-block:: python

            ['Flagged']
            ['Seen', 'Deleted']
            ['Junk']

        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).add_flags(
            messages=messages, flags=flags, folder=folder
        )

    @action
    def set_flags(
        self,
        messages: List[int],
        flags: Union[str, List[str]],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Set a set of flags to the specified set of message IDs.

        :param messages: List of message IDs.
        :param flags: List of flags to be added. Examples:

          .. code-block:: python

            ['Flagged']
            ['Seen', 'Deleted']
            ['Junk']

        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).set_flags(
            messages=messages, flags=flags, folder=folder
        )

    @action
    def remove_flags(
        self,
        messages: List[int],
        flags: Union[str, List[str]],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Remove a set of flags to the specified set of message IDs.

        :param messages: List of message IDs.
        :param flags: List of flags to be added. Examples:

          .. code-block:: python

            ['Flagged']
            ['Seen', 'Deleted']
            ['Junk']

        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).remove_flags(
            messages=messages, flags=flags, folder=folder
        )

    @action
    def flag_messages(
        self,
        messages: List[int],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Add a flag/star to the specified set of message IDs.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).flag_messages(
            messages=messages, folder=folder
        )

    @action
    def unflag_messages(
        self,
        messages: List[int],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Remove a flag/star from the specified set of message IDs.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).unflag_messages(
            messages=messages, folder=folder
        )

    @action
    def flag_message(
        self, message: int, folder: str = 'INBOX', account: Optional[AccountType] = None
    ):
        """
        Add a flag/star to the specified set of message ID.

        :param message: Message ID.
        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).flag_message(message=message, folder=folder)

    @action
    def unflag_message(
        self, message: int, folder: str = 'INBOX', account: Optional[AccountType] = None
    ):
        """
        Remove a flag/star from the specified set of message ID.

        :param message: Message ID.
        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).unflag_message(
            message=message, folder=folder
        )

    @action
    def delete_messages(
        self,
        messages: List[int],
        folder: str = 'INBOX',
        expunge: bool = True,
        account: Optional[AccountType] = None,
    ):
        """
        Set a specified set of message IDs as deleted.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param expunge: If set then the messages will also be expunged from the
            folder, otherwise they will only be marked as deleted (default:
            ``True``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).delete_messages(
            messages=messages, folder=folder, expunge=expunge
        )

    @action
    def restore_messages(
        self,
        messages: List[int],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Remove the ``Deleted`` flag from the specified set of message IDs.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).restore_messages(
            messages=messages, folder=folder
        )

    @action
    def copy_messages(
        self,
        messages: List[int],
        destination: str,
        source: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Copy a set of messages IDs from a folder to another.

        :param messages: List of message IDs.
        :param source: Source folder.
        :param destination: Destination folder.
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).copy_messages(
            messages=messages, dest_folder=destination, source_folder=source
        )

    @action
    def move_messages(
        self,
        messages: List[int],
        destination: str,
        source: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        Move a set of messages IDs from a folder to another.

        :param messages: List of message IDs.
        :param source: Source folder.
        :param destination: Destination folder.
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).move_messages(
            messages=messages, dest_folder=destination, source_folder=source
        )

    @action
    def expunge_messages(
        self,
        messages: List[int],
        folder: str = 'INBOX',
        account: Optional[AccountType] = None,
    ):
        """
        When ``messages`` is not set, remove all the messages from ``folder``
        marked as ``Deleted``.

        :param folder: IMAP folder (default: ``INBOX``).
        :param messages: List of message IDs to expunge (default: all those
            marked as ``Deleted``).
        :param account: Account name or index (default: default account).
        """
        return self._get_in_plugin(account).expunge_messages(
            folder=folder, messages=messages
        )

    @action
    def send(
        self,
        to: Union[str, List[str]],
        from_: Optional[str] = None,
        cc: Optional[Union[str, List[str]]] = None,
        bcc: Optional[Union[str, List[str]]] = None,
        subject: str = '',
        body: str = '',
        body_type: str = 'plain',
        attachments: Optional[List[str]] = None,
        headers: Optional[Dict[str, str]] = None,
        account: Optional[AccountType] = None,
        **kwargs,
    ):
        """
        Send an email through the specified SMTP sender.

        :param to: Receiver(s), as comma-separated strings or list.
        :param from_: Sender email address (``from`` is also supported).
        :param cc: Carbon-copy addresses, as comma-separated strings or list
        :param bcc: Blind carbon-copy addresses, as comma-separated strings or list
        :param subject: Mail subject.
        :param body: Mail body.
        :param body_type: Mail body type, as a subtype of ``text/`` (e.g. ``html``) (default: ``plain``).
        :param attachments: List of attachment files to send.
        :param headers: Key-value map of headers to be added.
        :param account: Account name/index to be used (default: default account).
        """
        plugin = self._get_out_plugin(account)
        headers = {k.lower(): v for k, v in (headers or {}).items()}
        cc = ([cc] if isinstance(cc, str) else cc) or []
        bcc = ([bcc] if isinstance(bcc, str) else bcc) or []
        return plugin.send(
            to=to,
            from_=(
                from_
                or kwargs.get('from')
                or (headers or {}).get('from')
                or plugin.account.display_name
                or plugin.account.username
            ),
            cc=cc,
            bcc=bcc,
            subject=subject,
            body=body,
            body_type=body_type,
            attachments=attachments,
            headers=headers,
        )

    @staticmethod
    def _process_account_changes(
        account: str, cur_status: AccountsStatus, new_status: AccountsStatus
    ) -> AccountFolderChanges:
        folders = new_status.get(account) or {}
        mail_flag_changed: Dict[str, Dict[MailFlagType, Dict[int, bool]]] = defaultdict(
            lambda: defaultdict(lambda: defaultdict(bool))
        )

        for folder, folder_status in folders.items():
            for flag, new_mail in folder_status.items():
                cur_mail = (cur_status.get(account) or {}).get(folder, {}).get(flag, {})
                cur_mail_keys = set(map(int, cur_mail.keys()))
                new_mail_keys = set(map(int, new_mail.keys()))
                mail_flag_added_keys = new_mail_keys - cur_mail_keys
                mail_flag_removed_keys = cur_mail_keys - new_mail_keys
                mail_flag_changed[folder][flag].update(
                    {
                        **{msg_id: True for msg_id in mail_flag_added_keys},
                        **{msg_id: False for msg_id in mail_flag_removed_keys},
                    }
                )

        return mail_flag_changed

    def _generate_account_events(
        self, account: str, msgs: Dict[int, Mail], folder_changes: AccountFolderChanges
    ):
        for folder, changes in folder_changes.items():
            for flag, flag_changes in changes.items():
                for msg_id, flag_added in flag_changes.items():
                    msg = msgs.get(msg_id)
                    if not msg:
                        continue

                    if flag == MailFlagType.UNREAD:
                        evt_type = UnseenMailEvent if flag_added else SeenMailEvent
                    elif flag == MailFlagType.FLAGGED:
                        evt_type = (
                            FlaggedMailEvent if flag_added else UnflaggedMailEvent
                        )
                    else:
                        continue

                    self._bus.post(
                        evt_type(
                            account=account,
                            folder=folder,
                            message=msg,
                        )
                    )

    def _generate_events(self, cur_status: AccountsStatus, new_status: AccountsStatus):
        for account in new_status:
            mail_flag_changed = self._process_account_changes(
                account, cur_status, new_status
            )

            msg_ids = {
                msg_id
                for _, changes in mail_flag_changed.items()
                for _, flag_changes in changes.items()
                for msg_id, _ in flag_changes.items()
            }

            if not msg_ids:
                continue

            acc = self._get_account(account)
            if not acc.incoming:
                continue

            msgs = acc.incoming.get_messages(*msg_ids, with_body=False)
            self._generate_account_events(account, msgs, mail_flag_changed)

    def _update_status(self, status: AccountsStatus):
        self._status = status
        with open(self._status_file, 'w') as f:
            status.write(f)

    def loop(self):
        cur_status = self._status.copy()
        new_status = self._check_mailboxes()
        self._generate_events(cur_status, new_status)
        self._update_status(new_status)

    def main(self):
        while not self.should_stop():
            try:
                self.loop()
            except Exception as e:
                self.logger.exception(e)
            finally:
                self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



from dataclasses import asdict, dataclass

from platypush.plugins.camera import CameraInfo, Camera


@dataclass
class PiCameraInfo(CameraInfo):
    """
    PiCamera info dataclass.
    """

    sharpness: int = 0
    contrast: int = 0
    brightness: int = 50
    video_stabilization: bool = False
    iso: int = 0
    exposure_compensation: int = 0
    hdr_mode: str = 'auto'
    meter_mode: str = 'average'
    awb_mode: str = 'auto'

    def to_dict(self) -> dict:
        return asdict(self)


class PiCamera(Camera):
    """
    PiCamera model.
    """

    info: PiCameraInfo  # type: ignore


# vim:sw=4:ts=4:et:



from contextlib import contextmanager
from email.message import Message
from typing import Dict, Generator

from smtplib import SMTP, SMTP_SSL

from .._model import TransportEncryption
from .._plugin import MailOutPlugin


class MailSmtpPlugin(MailOutPlugin):
    """
    Plugin to interact with a mail server over SMTP.
    """

    @classmethod
    def _matches_url_scheme(cls, scheme: str) -> bool:
        return scheme in ('smtp', 'smtps')

    @classmethod
    def default_ports(cls) -> Dict[TransportEncryption, int]:
        return {
            TransportEncryption.NONE: 25,
            TransportEncryption.SSL: 465,
            TransportEncryption.STARTTLS: 587,
        }

    @contextmanager
    def connect(self) -> Generator[SMTP, None, None]:
        smtp_args = {
            'host': self.server.server,
            'port': self.server.port,
        }

        if self.server.encryption == TransportEncryption.SSL:
            client_type = SMTP_SSL
            if self.server.certfile:
                smtp_args.update(certfile=self.server.certfile)
            if self.server.keyfile:
                smtp_args.update(keyfile=self.server.keyfile)
        else:
            client_type = SMTP

        client = client_type(**smtp_args)

        if self.server.encryption == TransportEncryption.STARTTLS:
            client.ehlo()
            client.starttls()
        else:
            client.ehlo_or_helo_if_needed()

        pwd = None
        try:
            pwd = self.account.get_password()
        except AssertionError:
            pass

        if pwd:
            client.login(self.account.username, pwd)

        yield client
        client.quit()

    def send_message(self, message: Message, **_):
        with self.connect() as client:
            errors = client.sendmail(
                message['From'], message['To'], message.as_string()
            )

        assert not errors, 'Failed to send message: ' + str(
            [f'{code}: {err}' for code, err in errors.items()]
        )


# vim:sw=4:ts=4:et:



import json

from flask import abort, request, Response, Blueprint
from markupsafe import escape

from platypush.backend.http.app import template_folder
from platypush.common.spotify import SpotifyMixin
from platypush.utils import get_redis

spotify = Blueprint('spotify', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    spotify,
]


@spotify.route('/spotify/auth_callback', methods=['GET'])
def auth_callback():
    """
    This route is used as a callback URL for Spotify API authentication flows.
    """

    code = request.args.get('code')
    state = request.args.get('state')
    error = request.args.get('error')

    if not state:
        abort(400, 'No state parameters provided')

    msg = {'error': code} if error else {'code': code}
    get_redis().rpush(SpotifyMixin.get_spotify_queue_for_state(state), json.dumps(msg))

    if error:
        return Response(f'Authentication failed: {escape(error)}')

    return Response('Authentication successful. You can now close this window')


# vim:sw=4:ts=4:et:



import pytest

from .utils import register_user, send_request as _send_request


@pytest.fixture(scope='module')
def expected_registration_redirect(base_url):
    yield '{base_url}/register?redirect={base_url}/execute'.format(base_url=base_url)


@pytest.fixture(scope='module')
def expected_login_redirect(base_url):
    yield '{base_url}/login?redirect={base_url}/execute'.format(base_url=base_url)


def send_request(**kwargs):
    return _send_request('shell.exec', args={'cmd': 'echo ping'}, **kwargs)


def test_request_with_no_registered_users(base_url, expected_registration_redirect):
    """
    An /execute request performed before any user is registered should redirect to the registration page.
    """
    response = send_request(authenticate=False, parse_json=False)
    assert response.status_code == 412, (
        'No users registered, but the execute endpoint returned '
        f'{response.status_code}'
    )


def test_first_user_registration(base_url):
    """
    Emulate a first user registration through form and get the session_token.
    """
    response = register_user()

    assert len(response.history) > 0, 'Redirect missing from the history'
    assert (
        'session_token' in response.history[0].cookies
    ), 'No session_token returned upon registration'
    assert (
        '{base_url}/'.format(base_url=base_url) == response.url
    ), 'The registration form did not redirect to the main panel'


def test_unauthorized_request_with_registered_user(base_url, expected_login_redirect):
    """
    After a first user has been registered any unauthenticated call to /execute should redirect to /login.
    """
    response = send_request(authenticate=False, parse_json=False)
    assert response.status_code == 401, (
        'An unauthenticated request after user registration should result in a '
        f'401 error, got {response.status_code} instead'
    )


def test_authorized_request_with_registered_user(base_url):
    # A request authenticated with user/pass should succeed.
    response = send_request(authenticate=True)
    assert (
        response.output.strip() == 'ping'
    ), 'The request did not return the expected output'


def test_request_with_wrong_credentials(base_url, expected_login_redirect):
    # A request with the wrong user/pass should fail.
    response = send_request(
        authenticate=False, auth=('wrong', 'wrong'), parse_json=False
    )
    assert response.status_code == 401, (
        'A request with wrong credentials should fail with status 401, '
        f'got {response.status_code} instead'
    )


if __name__ == '__main__':
    pytest.main()


# vim:sw=4:ts=4:et:



from typing import Dict, List

from platypush.common.sensors import Numeric
from platypush.entities.devices import Device
from platypush.entities.distance import DistanceSensor
from platypush.entities.illuminance import IlluminanceSensor
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


# pylint: disable=too-many-ancestors
class SensorLtr559Plugin(SensorPlugin):
    """
    Plugin to interact with an `LTR559 <https://shop.pimoroni.com/products/ltr-559-light-proximity-sensor-breakout>`_
    light and proximity sensor.
    """

    def __init__(self, **kwargs):
        import ltr559

        super().__init__(**kwargs)
        self.ltr = ltr559.LTR559()

    @action
    def get_measurement(self, *_, **__):
        """
        :returns: dict. Example:

          .. code-block:: python

             output = {
                 "light": 109.3543,     # Lux
                 "proximity": 103       # The higher the value, the nearest the object, within a ~5cm range
             }

        """
        self.ltr.update_sensor()
        return {
            'light': self.ltr.get_lux(),
            'proximity': self.ltr.get_proximity(),
        }

    def transform_entities(self, entities: Dict[str, Numeric]) -> List[Device]:
        sensors = []

        if entities.get('light') is not None:
            sensors.append(
                IlluminanceSensor(
                    id='ltr559:illuminance',
                    name='illuminance',
                    value=entities['light'],
                    unit='lux',
                )
            )

        if entities.get('proximity') is not None:
            sensors.append(
                DistanceSensor(
                    id='ltr559:proximity',
                    name='proximity',
                    value=entities['proximity'],
                    unit='mm',
                )
            )

        if not sensors:
            return []

        return [
            Device(
                id='ltr559',
                name='LTR559 Sensor',
                children=sensors,
            )
        ]


# vim:sw=4:ts=4:et:



from .action import Action
from .argument import Argument
from .constructor import Constructor
from .integration import Integration
from .message import Message
from .returns import ReturnValue


__all__ = [
    "Action",
    "Argument",
    "Constructor",
    "Integration",
    "Message",
    "ReturnValue",
]



from sqlalchemy import Column, Integer, ForeignKey

from platypush.common.db import is_defined

from .sensors import NumericSensor


if not is_defined('temperature_sensor'):

    class TemperatureSensor(NumericSensor):
        __tablename__ = 'temperature_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from marshmallow import EXCLUDE, fields
from marshmallow.schema import Schema
from marshmallow.validate import OneOf

from platypush.schemas import StrippedString


class PiholeStatusSchema(Schema):
    """
    Schema for a Pi-hole status response.


    "output": {
      "server": "dns.fabiomanganiello.com",
      "status": "enabled",
      "ads_percentage": 6.7,
      "blocked": 37191,
      "cached": 361426,
      "domain_count": 1656690,
      "forwarded": 150187,
      "queries": 552076,
      "total_clients": 57,
      "total_queries": 552076,
      "unique_clients": 41,
      "unique_domains": 39348,
      "version": "5.18.2"
    },
    """

    class Meta:  # type: ignore
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    server = StrippedString(
        required=True,
        metadata={
            'description': 'Hostname or IP of the Pi-hole server',
            'example': '192.168.1.254',
        },
    )

    status = fields.String(
        required=True,
        validate=OneOf(['enabled', 'disabled']),
        metadata={
            'description': 'Status of the Pi-hole server',
            'example': 'enabled',
        },
    )

    ads_percentage = fields.Float(
        required=True,
        metadata={
            'description': 'Percentage of ads blocked by the Pi-hole server',
            'example': 6.7,
        },
    )

    blocked = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of blocked queries',
            'example': 37191,
        },
    )

    cached = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of cached queries',
            'example': 361426,
        },
    )

    domain_count = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of domains resolved the Pi-hole server',
            'example': 1656690,
        },
    )

    forwarded = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of forwarded queries',
            'example': 150187,
        },
    )

    queries = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of processed queries since the latest restart',
            'example': 552076,
        },
    )

    total_clients = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of connected clients',
            'example': 57,
        },
    )

    total_queries = fields.Integer(
        required=True,
        metadata={
            'description': 'Total number of queries processed by the Pi-hole server',
            'example': 552076,
        },
    )

    unique_clients = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of unique IP addresses connected to the Pi-hole server',
            'example': 41,
        },
    )

    unique_domains = fields.Integer(
        required=True,
        metadata={
            'description': 'Number of unique domains resolved by the Pi-hole server',
            'example': 39348,
        },
    )

    version = StrippedString(
        required=True,
        metadata={
            'description': 'Version of the Pi-hole server',
            'example': '5.18.2',
        },
    )
